#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

"""
This module is intended for general purpose functions that can be used
thoughout the qiling framework
"""
import importlib, os, copy, re, pefile, configparser, logging
from logging import LogRecord
from pathlib import Path

from unicorn import UcError, UC_ERR_READ_UNMAPPED, UC_ERR_FETCH_UNMAPPED
from keystone import *
from capstone import *

from .exception import *
from .const import QL_ARCH, QL_ARCH_ALL, QL_ENDIAN, QL_OS, QL_OS_ALL, QL_OUTPUT, QL_DEBUGGER, QL_ARCH_32BIT, QL_ARCH_64BIT, QL_ARCH_16BIT
from .const import debugger_map, arch_map, os_map

FMT_STR = "%(levelname)s\t%(message)s"

# \033 -> ESC
# ESC [ -> CSI
# CSI %d;%d;... m -> SGR
class COLOR_CODE:
    WHITE = '\033[37m'
    CRIMSON = '\033[31m'
    RED = '\033[91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    MAGENTA = '\033[95m'
    CYAN = '\033[96m'
    ENDC = '\033[0m'

LEVEL_COLORS = {
    'WARNING': COLOR_CODE.YELLOW,
    'INFO': COLOR_CODE.BLUE,
    'DEBUG': COLOR_CODE.MAGENTA,
    'CRITICAL': COLOR_CODE.CRIMSON,
    'ERROR': COLOR_CODE.RED
}
class QilingColoredFormatter(logging.Formatter):
    def __init__(self, ql, *args, **kwargs):
        super(QilingColoredFormatter, self).__init__(*args, **kwargs)
        self._ql = ql

    def get_colored_level(self, record: LogRecord):
        LEVEL_NAME = {
            'WARNING': f"{COLOR_CODE.YELLOW}[!]{COLOR_CODE.ENDC}",
            'INFO': f"{COLOR_CODE.BLUE}[=]{COLOR_CODE.ENDC}",
            'DEBUG': f"{COLOR_CODE.MAGENTA}[+]{COLOR_CODE.ENDC}",
            'CRITICAL': f"{COLOR_CODE.CRIMSON}[x]{COLOR_CODE.ENDC}",
            'ERROR': f"{COLOR_CODE.RED}[x]{COLOR_CODE.ENDC}"
        }
        return LEVEL_NAME[record.levelname]

    def format(self, record: LogRecord):
        record.levelname = self.get_colored_level(record)
        try:
            cur_thread = self._ql.os.thread_management.cur_thread
            if cur_thread is not None:
                record.levelname = f"{record.levelname} {COLOR_CODE.GREEN}{str(cur_thread)}{COLOR_CODE.ENDC}"
        except AttributeError:
            pass
        return super(QilingColoredFormatter, self).format(record)

class QilingPlainFormatter(logging.Formatter):
    def __init__(self, ql, *args, **kwargs):
        super(QilingPlainFormatter, self).__init__(*args, **kwargs)
        self._ql = ql
    
    def get_level(self, record: LogRecord):
        LEVEL_NAME = {
            'WARNING': "[!]",
            'INFO': "[=]",
            'DEBUG': "[+]",
            'CRITICAL': "[x]",
            'ERROR': "[x]"
        }
        return LEVEL_NAME[record.levelname]

    def format(self, record: LogRecord):
        record.levelname = self.get_level(record)
        try:
            cur_thread = self._ql.os.thread_management.cur_thread
            if cur_thread is not None:
                record.levelname = f"{record.levelname} {str(cur_thread)}"
        except AttributeError:
            pass
        return super(QilingPlainFormatter, self).format(record)

class RegexFilter(logging.Filter):
    def __init__(self, filters):
        super(RegexFilter, self).__init__()
        self.update_filters(filters)
    
    def update_filters(self, filters):
        self._filters = [ re.compile(ft) for ft in  filters ]

    def filter(self, record: LogRecord):
        msg = record.getMessage()
        for ft in self._filters:
            if re.match(ft, msg):
                return True
        return False

class QlFileDes:
    def __init__(self, init):
        self.__fds = init

    def __getitem__(self, idx):
        return self.__fds[idx]

    def __setitem__(self, idx, val):
        self.__fds[idx] = val

    def __iter__(self):
        return iter(self.__fds)

    def __repr__(self):
        return repr(self.__fds)
    
    def save(self):
        return self.__fds

    def restore(self, fds):
        self.__fds = fds


class QlStopOptions(object):
    def __init__(self, stackpointer=False, exit_trap=False):
        super().__init__()
        self._stackpointer = stackpointer
        self._exit_trap = exit_trap

    @property
    def stackpointer(self) -> bool:
        return self._stackpointer

    @property
    def exit_trap(self) -> bool:
        return self._exit_trap

    @property
    def any(self) -> bool:
        return self.stackpointer or self.exit_trap


def catch_KeyboardInterrupt(ql):
    def decorator(func):
        def wrapper(*args, **kw):
            try:
                return func(*args, **kw)
            except BaseException as e:
                from .os.const import THREAD_EVENT_UNEXECPT_EVENT
                ql.os.stop()
                ql._internal_exception = e
        return wrapper
    return decorator

def ql_get_arch_bits(arch):
    if arch in QL_ARCH_16BIT:
        return 16
    if arch in QL_ARCH_32BIT:
        return 32
    if arch in QL_ARCH_64BIT:
        return 64
    raise QlErrorArch("Invalid Arch")

def ql_is_valid_ostype(ostype):
    if ostype not in QL_OS_ALL:
        return False
    return True

def ql_is_valid_arch(arch):
    if arch not in QL_ARCH_ALL:
        return False
    return True

def loadertype_convert_str(ostype):
    adapter = {
        QL_OS.LINUX: "ELF",
        QL_OS.MACOS: "MACHO",
        QL_OS.FREEBSD: "ELF",
        QL_OS.WINDOWS: "PE",
        QL_OS.UEFI: "PE_UEFI",
        QL_OS.DOS: "DOS",
    }
    return adapter.get(ostype)

def ostype_convert_str(ostype):
    adapter = {}
    adapter.update(os_map)
    adapter = {v: k for k, v in adapter.items()}
    return adapter.get(ostype)

def ostype_convert(ostype):
    # this is for ql.platform
    if ostype == "darwin":
        ostype = "macos"
    adapter = {}
    adapter.update(os_map)
    if ostype in adapter:
        return adapter[ostype]
    # invalid
    return None

def arch_convert_str(arch):
    adapter = {}
    adapter.update(arch_map)
    adapter = {v: k for k, v in adapter.items()}
    return adapter.get(arch)

def arch_convert(arch):
    adapter = {}
    adapter.update(arch_map)
    if arch in adapter:
        return adapter[arch]
    # invalid
    return None

def output_convert(output):
    adapter = {
        None: QL_OUTPUT.DEFAULT,
        "default": QL_OUTPUT.DEFAULT,
        "disasm": QL_OUTPUT.DISASM,
        "debug": QL_OUTPUT.DEBUG,
        "dump": QL_OUTPUT.DUMP,
    }
    if output in adapter:
        return adapter[output]
    # invalid
    return QL_OUTPUT.DEFAULT

def debugger_convert(debugger):
    adapter = {}
    adapter.update(debugger_map)
    if debugger in adapter:
        return adapter[debugger]
    # invalid
    return None, None

def debugger_convert_str(debugger_id):
    adapter = {}
    adapter.update(debugger_map)
    adapter = {v: k for k, v in adapter.items()}
    if debugger_id in adapter:
        return adapter[debugger_id]
    # invalid
    return None, None

# Call `function_name` in `module_name`.
# e.g. map_syscall in qiling.os.linux.map_syscall
def ql_get_module_function(module_name, function_name = None):
    
    try:
        imp_module = importlib.import_module(module_name)
    except Exception as ex:
        raise QlErrorModuleNotFound("Unable to import module %s (%s)" % (module_name, ex))

    try:
        module_function = getattr(imp_module, function_name)
    except:
        raise QlErrorModuleFunctionNotFound("Unable to import %s from %s" % (function_name, imp_module))

    return module_function

def ql_elf_parse_emu_env(path):
    def getident():
        return elfdata

    with open(path, "rb") as f:
        elfdata = f.read()[:20]

    ident = getident()
    ostype = None
    arch = None
    archendian = None

    if ident[: 4] == b'\x7fELF':
        elfbit = ident[0x4]
        endian = ident[0x5]
        osabi = ident[0x7]
        e_machine = ident[0x12:0x14]

        if osabi == 0x09:
            ostype = QL_OS.FREEBSD
        elif osabi in (0x0, 0x03) or osabi >= 0x11:
            ostype = QL_OS.LINUX

        if e_machine == b"\x03\x00":
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.X86
        elif e_machine == b"\x08\x00" and endian == 1 and elfbit == 1:
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.MIPS
        elif e_machine == b"\x00\x08" and endian == 2 and elfbit == 1:
            archendian = QL_ENDIAN.EB
            arch = QL_ARCH.MIPS
        elif e_machine == b"\x28\x00" and endian == 1 and elfbit == 1:
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.ARM
        elif e_machine == b"\x00\x28" and endian == 2 and elfbit == 1:
            archendian = QL_ENDIAN.EB
            arch = QL_ARCH.ARM            
        elif e_machine == b"\xB7\x00":
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.ARM64
        elif e_machine == b"\x3E\x00":
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.X8664
        else:
            arch = None

    return arch, ostype, archendian

def ql_macho_parse_emu_env(path):
   
    def getident():
        return machodata

    with open(path, "rb") as f:
        machodata = f.read()[:32]

    ident = getident()

    macho_macos_sig64 = b'\xcf\xfa\xed\xfe'
    macho_macos_sig32 = b'\xce\xfa\xed\xfe'
    macho_macos_fat = b'\xca\xfe\xba\xbe'  # should be header for FAT

    ostype = None
    arch = None
    archendian = None

    if ident[: 4] in (macho_macos_sig32, macho_macos_sig64, macho_macos_fat):
        ostype = QL_OS.MACOS
    else:
        ostype = None

    if ostype:
        # if ident[0x7] == 0: # 32 bit
        #    arch = QL_ARCH.X86
        if ident[0x4] == 7 and ident[0x7] == 1:  # X86 64 bit
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.X8664
        elif ident[0x4] == 12 and ident[0x7] == 1:  # ARM64  ident[0x4] = 0x0C
            archendian = QL_ENDIAN.EL
            arch = QL_ARCH.ARM64
        else:
            arch = None

    return arch, ostype, archendian


def ql_pe_parse_emu_env(path):
    try:
        pe = pefile.PE(path, fast_load=True)
    except:
        return None, None, None

    ostype = None
    arch = None
    archendian = None

    machine_map = {
        pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_I386']: QL_ARCH.X86,
        pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_AMD64']: QL_ARCH.X8664,
        pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_ARM']: QL_ARCH.ARM,
        pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_THUMB']: QL_ARCH.ARM,
        # pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_ARM64']     :   QL_ARCH.ARM64       #pefile does not have the definition
        # for IMAGE_FILE_MACHINE_ARM64
        0xAA64: QL_ARCH.ARM64  # Temporary workaround for Issues #21 till pefile gets updated
    }
    # get arch
    archendian = QL_ENDIAN.EL
    arch = machine_map.get(pe.FILE_HEADER.Machine)

    if arch:
        if pe.OPTIONAL_HEADER.Subsystem >= pefile.SUBSYSTEM_TYPE['IMAGE_SUBSYSTEM_EFI_APPLICATION'] and \
        pe.OPTIONAL_HEADER.Subsystem <= pefile.SUBSYSTEM_TYPE['IMAGE_SUBSYSTEM_EFI_ROM'] :
            ostype = QL_OS.UEFI
        else:
            ostype = QL_OS.WINDOWS
    else:
        ostype = None

    return arch, ostype, archendian


def ql_guess_emu_env(path):
    arch = None
    ostype = None
    archendian = None

    if os.path.isdir(path) and (str(path)).endswith(".kext"):
        return QL_ARCH.X8664, QL_OS.MACOS, QL_ENDIAN.EL

    if os.path.isfile(path) and (str(path)).endswith(".DOS_COM"):
        return QL_ARCH.A8086, QL_OS.DOS, QL_ENDIAN.EL

    if os.path.isfile(path) and (str(path)).endswith(".DOS_MBR"):
        return QL_ARCH.A8086, QL_OS.DOS, QL_ENDIAN.EL

    if os.path.isfile(path) and (str(path)).endswith(".DOS_EXE"):
        return QL_ARCH.A8086, QL_OS.DOS, QL_ENDIAN.EL

    arch, ostype, archendian = ql_elf_parse_emu_env(path)

    if arch == None or ostype == None or archendian == None:
        arch, ostype, archendian = ql_macho_parse_emu_env(path)

    if arch == None or ostype == None or archendian == None:
        arch, ostype, archendian = ql_pe_parse_emu_env(path)
  
    if ostype not in (QL_OS_ALL):
        raise QlErrorOsType("File does not belong to either 'linux', 'windows', 'freebsd', 'macos', 'ios', 'dos'")

    return arch, ostype, archendian


def loader_setup(ostype, ql):
    loadertype_str = loadertype_convert_str(ostype)
    function_name = "QlLoader" + loadertype_str
    return ql_get_module_function(f"qiling.loader.{loadertype_str.lower()}", function_name)(ql)


def component_setup(component_type, component_name, ql):
    function_name = "Ql" + component_name.capitalize() + "Manager"
    return ql_get_module_function(f"qiling.{component_type}.{component_name}", function_name)(ql)


def debugger_setup(debugger, ql):
    # default remote server
    remotedebugsrv = "gdb"
    debug_opts = [None, None]

    if debugger != True and type(debugger) == str:      
        debug_opts = debugger.split(":")

        if len(debug_opts) == 2 and debug_opts[0] != "qdb":
            pass
        else:  
            remotedebugsrv, *debug_opts = debug_opts
            
        
        if debugger_convert(remotedebugsrv) not in (QL_DEBUGGER):
            raise QlErrorOutput("Error: Debugger not supported")
        
    debugsession = ql_get_module_function(f"qiling.debugger.{remotedebugsrv}.{remotedebugsrv}", f"Ql{str.capitalize(remotedebugsrv)}")

    return debugsession(ql, *debug_opts)

def arch_setup(archtype, ql):
    if not ql_is_valid_arch(archtype):
        raise QlErrorArch("Invalid Arch")
    
    if archtype == QL_ARCH.ARM_THUMB:
        archtype =  QL_ARCH.ARM

    archmanager = arch_convert_str(archtype).upper()
    archmanager = ("QlArch" + archmanager)

    if archtype == QL_ARCH.X8664:
        arch_str = "x86"
    else:
        arch_str = arch_convert_str(archtype)

    return ql_get_module_function(f"qiling.arch.{arch_str.lower()}", archmanager)(ql)


# This function is extracted from os_setup so I put it here.
def ql_syscall_mapping_function(ostype):
    ostype_str = ostype_convert_str(ostype)
    return ql_get_module_function(f"qiling.os.{ostype_str.lower()}.map_syscall", "map_syscall")


def os_setup(archtype, ostype, ql):
    if not ql_is_valid_ostype(ostype):
        raise QlErrorOsType("Invalid OSType")

    if not ql_is_valid_arch(archtype):
        raise QlErrorArch("Invalid Arch %s" % archtype)

    ostype_str = ostype_convert_str(ostype)
    ostype_str = ostype_str.capitalize()
    function_name = "QlOs" + ostype_str
    return ql_get_module_function(f"qiling.os.{ostype_str.lower()}.{ostype_str.lower()}", function_name)(ql)


def profile_setup(ostype, profile, ql):
    _profile = "Default"
    if profile != None:
        _profile = profile
        
    debugmsg = "Profile: %s" % _profile

    os_profile = os.path.join(os.path.dirname(os.path.abspath(__file__)), "profiles", ostype_convert_str(ostype) + ".ql")

    if profile:
        profiles = [os_profile, profile]
    else:
        profiles = [os_profile]

    config = configparser.ConfigParser()
    config.read(profiles)
    return config, debugmsg

def ql_resolve_logger_level(output, verbose):
    level = logging.INFO
    if output in (QL_OUTPUT.DEBUG, QL_OUTPUT.DUMP, QL_OUTPUT.DISASM):
        level = logging.DEBUG
    else:
        if verbose == 0:
            level = logging.WARNING
        elif verbose >= 4:
            level = logging.DEBUG
        elif verbose >= 1:
            level = logging.INFO
    
    return level

QL_INSTANCE_ID = 114514

# TODO: qltool compatibility
def ql_setup_logger(ql, log_file, console, filters, multithread, log_override):
    global QL_INSTANCE_ID

    # If there is an override for our logger, then use it.
    if log_override is not None:
        log = log_override
    else:
        # We should leave the root logger untouched.
        log = logging.getLogger(f"qiling{QL_INSTANCE_ID}")
        QL_INSTANCE_ID += 1
        
        # Disable propagation to avoid duplicate output.
        log.propagate = False
        # Clear all handlers and filters.
        log.handlers = []
        log.filters = []    

        # Do we have console output?
        if console:
            handler = logging.StreamHandler()
            formatter = QilingColoredFormatter(ql, FMT_STR)
            handler.setFormatter(formatter)
            log.addHandler(handler)
        else:
            log.setLevel(logging.CRITICAL)

        # Do we have to write log to a file?
        if log_file is not None:
            handler = logging.FileHandler(log_file)
            formatter = QilingPlainFormatter(ql, FMT_STR)
            handler.setFormatter(formatter)
            log.addHandler(handler)

    # Remeber to add filters if necessary.
    # If there aren't any filters, we do add the filters until users specify any.
    log_filter = None

    if filters is not None and type(filters) == list and len(filters) != 0:
        log_filter = RegexFilter(filters)
        log.addFilter(log_filter)
    
    log.setLevel(logging.INFO)

    return log, log_filter


# verify if emulator returns properly
def verify_ret(ql, err):
    ql.log.debug("Got exception %u: init SP = %x, current SP = %x, PC = %x" %(err.errno, ql.os.init_sp, ql.reg.arch_sp, ql.reg.arch_pc))
    # print("Got exception %u: init SP = %x, current SP = %x, PC = %x" %(err.errno, ql.os.init_sp, self.reg.arch_sp, self.reg.arch_pc))

    ql.os.RUN = False

    # timeout is acceptable in this case
    if err.errno in (UC_ERR_READ_UNMAPPED, UC_ERR_FETCH_UNMAPPED):
        if ql.ostype == QL_OS.MACOS:
            if ql.loader.kext_name:
                # FIXME: Should I push saved RIP before every method callings of IOKit object?
                if ql.os.init_sp == ql.reg.arch_sp - 8:
                    pass
                else:
                    raise
        
        if ql.archtype == QL_ARCH.X8664: # Win64
            if ql.os.init_sp == ql.reg.arch_sp or ql.os.init_sp + 8 == ql.reg.arch_sp or ql.os.init_sp + 0x10 == ql.reg.arch_sp:  # FIXME
                # 0x11626	 c3	  	ret
                # print("OK, stack balanced!")
                pass
            else:
                raise
        else:   # Win32
            if ql.os.init_sp + 12 == ql.reg.arch_sp:   # 12 = 8 + 4
                # 0x114dd	 c2 08 00	  	ret 	8
                pass
            else:
                raise
    else:
        raise        



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import ctypes, struct

from enum import IntEnum


class POINTER64(ctypes.Structure):
    _fields_ = [("value", ctypes.c_uint64)]

class list_entry(ctypes.Structure):
    _fields_ = (
        ("le_next", POINTER64),
        ("le_prev", POINTER64),
    )

class list_head(ctypes.Structure):
    _fields_ = (
        ("lh_first", POINTER64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

class tailq_head(ctypes.Structure):
    _fields_ = (
        ("tqh_first", POINTER64),
        ("tqh_last", POINTER64),
    )

class slist_head(ctypes.Structure):
    _fields_ = (
        ("slh_first", POINTER64),
    )

class tailq_entry(ctypes.Structure):
    _fields_ = (
        ("tqe_next", POINTER64),
        ("tqe_prev", POINTER64),
    )

# struct IOExternalMethodArguments
# {
#     uint32_t      version;

#     uint32_t      selector;

#     mach_port_t           asyncWakePort;
#     io_user_reference_t * asyncReference;
#     uint32_t              asyncReferenceCount;

#     const uint64_t *    scalarInput;
#     uint32_t      scalarInputCount;

#     const void *  structureInput;
#     uint32_t      structureInputSize;

#     IOMemoryDescriptor * structureInputDescriptor;
   
#     uint64_t *        scalarOutput;
#     uint32_t      scalarOutputCount;

#     void *        structureOutput;
#     uint32_t      structureOutputSize;

#     IOMemoryDescriptor * structureOutputDescriptor;
#     uint32_t       structureOutputDescriptorSize;

#     uint32_t      __reservedA;

#     OSObject **         structureVariableOutputData;

#     uint32_t      __reserved[30];
# };

class IOExternalMethodArguments(ctypes.Structure):
    _pack_ = 8
    _fields_ = (
        ("_version", ctypes.c_uint32),
        ("_selector", ctypes.c_uint32),
        ("_asyncWakePort", ctypes.c_uint32),
        ("_asyncReference", POINTER64),
        ("_asyncReferenceCount", ctypes.c_uint32),
        ("_scalarInput", POINTER64),
        ("_scalarInputCount", ctypes.c_uint32),
        ("_structureInput", POINTER64),
        ("_structureInputSize", ctypes.c_uint32),
        ("_structureInputDescriptor", POINTER64),
        ("_scalarOutput", POINTER64),
        ("_scalarOutputCount", ctypes.c_uint32),
        ("_structureOutput", POINTER64),
        ("_structureOutputSize", ctypes.c_uint32),
        ("_structureOutputDescriptor", POINTER64),
        ("_structureOutputDescriptorSize", ctypes.c_uint32),
        ("___reservedA", ctypes.c_uint32),
        ("_structureVariableOutputData", POINTER64),
        ("___reserved", ctypes.c_uint32 * 30)
    )
    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    @property
    def scalarInput(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        obj = type(self).from_buffer(data)
        array = self.ql.mem.read(obj._scalarInput.value, obj._scalarInputCount)
        return [struct.unpack("<Q", array[i: i + 8]) for i in range(0, len(array), 8)]

    @property
    def structureInput(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        obj = type(self).from_buffer(data)
        return self.ql.mem.read(obj._structureInput.value, obj._structureInputSize)

    @property
    def scalarOutput(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        obj = type(self).from_buffer(data)
        array = self.ql.mem.read(obj._scalarOutput.value, obj._scalarOutputCount)
        return [struct.unpack("<Q", array[i: i + 8]) for i in range(0, len(array), 8)]

    @property
    def structureOutput(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        obj = type(self).from_buffer(data)
        return self.ql.mem.read(obj._structureOutput.value, obj._structureOutputSize)

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# typedef IOReturn (*IOExternalMethodAction)(OSObject * target, void * reference, 
#                         IOExternalMethodArguments * arguments);
# struct IOExternalMethodDispatch
# {
#     IOExternalMethodAction function;
#     uint32_t           checkScalarInputCount;
#     uint32_t           checkStructureInputSize;
#     uint32_t           checkScalarOutputCount;
#     uint32_t           checkStructureOutputSize;
# };

class IOExternalMethodDispatch(ctypes.Structure):
    _pack_ = 8
    _fields_ = (
        ("function", POINTER64),
        ("checkScalarInputCount", ctypes.c_uint32),
        ("checkStructureInputSize", ctypes.c_uint32),
        ("checkScalarOutputCount", ctypes.c_uint32),
        ("checkStructureOutputSize", ctypes.c_uint32)
    )
    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# typedef struct kmod_info {
#     struct kmod_info  * next;
#     int32_t             info_version;           // version of this structure
#     uint32_t            id;
#     char                name[KMOD_MAX_NAME];
#     char                version[KMOD_MAX_NAME];
#     int32_t             reference_count;        // # linkage refs to this 
#     kmod_reference_t  * reference_list;         // who this refs (links on)
#     vm_address_t        address;                // starting address
#     vm_size_t           size;                   // total size
#     vm_size_t           hdr_size;               // unwired hdr size
#     kmod_start_func_t * start;
#     kmod_stop_func_t  * stop;
# } kmod_info_t;

class kmod_info_t(ctypes.Structure):
    _pack_ = 8
    _fields_ = (
        ("next", POINTER64),
        ("info_version", ctypes.c_int32),
        ("id", ctypes.c_uint32),
        ("name", ctypes.c_char * 64),
        ("version", ctypes.c_char * 64),
        ("reference_count", ctypes.c_int32),
        ("reference_list", POINTER64),
        ("address", POINTER64),
        ("size", ctypes.c_uint64),
        ("hdr_size", ctypes.c_uint64),
        ("start", POINTER64),
        ("stop", POINTER64)
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct queue_entry {
#     struct queue_entry  *next;      /* next element */
#     struct queue_entry  *prev;      /* previous element */
# };

class queue_entry(ctypes.Structure):
    _pack_ = 8
    _fields_ = (
        ("next", POINTER64),
        ("prev", POINTER64)
    )

# typedef struct _OSMallocTag_ {
#     queue_chain_t   OSMT_link;
#     uint32_t        OSMT_refcnt;
#     uint32_t        OSMT_state;
#     uint32_t        OSMT_attr;
#     char            OSMT_name[OSMT_MAX_NAME];
# }

class OSMallocTag(ctypes.Structure):
    _pack_ = 8
    _fields_ = (
        ("OSMT_link", queue_entry),
        ("OSMT_refcnt", ctypes.c_uint32),
        ("OSMT_state", ctypes.c_uint32),
        ("OSMT_attr", ctypes.c_uint32),
        ("OSMT_name", ctypes.c_char * 64)
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# typedef struct {
#     uint64_t            lck_grp_spin_util_cnt;
#     uint64_t            lck_grp_spin_held_cnt;
#     uint64_t            lck_grp_spin_miss_cnt;
#     uint64_t            lck_grp_spin_held_max;
#     uint64_t            lck_grp_spin_held_cum;
# } lck_grp_spin_stat_t;

# typedef struct {
#     uint64_t            lck_grp_mtx_util_cnt;
#     /* On x86, this is used as the "direct wait" count */
#     uint64_t            lck_grp_mtx_held_cnt;
#     uint64_t            lck_grp_mtx_miss_cnt;
#     uint64_t            lck_grp_mtx_wait_cnt;
#     /* Rest currently unused */
#     uint64_t            lck_grp_mtx_held_max;
#     uint64_t            lck_grp_mtx_held_cum;
#     uint64_t            lck_grp_mtx_wait_max;
#     uint64_t            lck_grp_mtx_wait_cum;
# } lck_grp_mtx_stat_t;

# typedef struct {
#     uint64_t            lck_grp_rw_util_cnt;
#     uint64_t            lck_grp_rw_held_cnt;
#     uint64_t            lck_grp_rw_miss_cnt;
#     uint64_t            lck_grp_rw_wait_cnt;
#     uint64_t            lck_grp_rw_held_max;
#     uint64_t            lck_grp_rw_held_cum;
#     uint64_t            lck_grp_rw_wait_max;
#     uint64_t            lck_grp_rw_wait_cum;
# } lck_grp_rw_stat_t;

# typedef struct _lck_grp_stat_ {
#     lck_grp_spin_stat_t lck_grp_spin_stat;
#     lck_grp_mtx_stat_t  lck_grp_mtx_stat;
#     lck_grp_rw_stat_t   lck_grp_rw_stat;
# } lck_grp_stat_t;

# typedef struct _lck_grp_ {
#     queue_chain_t       lck_grp_link;
#     uint32_t        lck_grp_refcnt;
#     uint32_t        lck_grp_spincnt;
#     uint32_t        lck_grp_mtxcnt;
#     uint32_t        lck_grp_rwcnt;
#     uint32_t        lck_grp_attr;
#     char            lck_grp_name[LCK_GRP_MAX_NAME];
#     lck_grp_stat_t      lck_grp_stat;
# } lck_grp_t;

class lck_grp_spin_stat_t(ctypes.Structure):
    _fields_ = (
        ("lck_grp_spin_util_cnt", ctypes.c_uint64),
        ("lck_grp_spin_held_cnt", ctypes.c_uint64),
        ("lck_grp_spin_miss_cnt", ctypes.c_uint64),
        ("lck_grp_spin_held_max", ctypes.c_uint64),
        ("lck_grp_spin_held_cum", ctypes.c_uint64),
    )

class lck_grp_mtx_stat_t(ctypes.Structure):
    _fields_ = (
        ("lck_grp_mtx_util_cnt", ctypes.c_uint64),
        ("lck_grp_mtx_held_cnt", ctypes.c_uint64),
        ("lck_grp_mtx_miss_cnt", ctypes.c_uint64),
        ("lck_grp_mtx_wait_cnt", ctypes.c_uint64),
        ("lck_grp_mtx_held_max", ctypes.c_uint64),
        ("lck_grp_mtx_held_cum", ctypes.c_uint64),
        ("lck_grp_mtx_wait_max", ctypes.c_uint64),
        ("lck_grp_mtx_wait_cum", ctypes.c_uint64),
    )

class lck_grp_rw_stat_t(ctypes.Structure):
    _fields_ = (
        ("lck_grp_rw_util_cnt", ctypes.c_uint64),
        ("lck_grp_rw_held_cnt", ctypes.c_uint64),
        ("lck_grp_rw_miss_cnt", ctypes.c_uint64),
        ("lck_grp_rw_wait_cnt", ctypes.c_uint64),
        ("lck_grp_rw_held_max", ctypes.c_uint64),
        ("lck_grp_rw_held_cum", ctypes.c_uint64),
        ("lck_grp_rw_wait_max", ctypes.c_uint64),
        ("lck_grp_rw_wait_cum", ctypes.c_uint64),
    )

class lck_grp_stat_t(ctypes.Structure):
    _fields_ = (
        ("lck_grp_spin_stat", lck_grp_spin_stat_t),
        ("lck_grp_mtx_stat", lck_grp_mtx_stat_t),
        ("lck_grp_rw_stat", lck_grp_rw_stat_t),
    )

class lck_grp_t(ctypes.Structure):
    _fields_ = (
        ("lck_grp_link", queue_entry),
        ("lck_grp_refcnt", ctypes.c_uint32),
        ("lck_grp_spincnt", ctypes.c_uint32),
        ("lck_grp_mtxcnt", ctypes.c_uint32),
        ("lck_grp_rwcnt", ctypes.c_uint32),
        ("lck_grp_attr", ctypes.c_uint32),
        ("lck_grp_name", ctypes.c_char * 64),
        ("lck_grp_stat", lck_grp_stat_t)
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# typedef struct _lck_grp_attr_ {
#     uint32_t    grp_attr_val;
# } lck_grp_attr_t;

class lck_grp_attr_t(ctypes.Structure):
    _fields_ = (
        ("grp_attr_val", ctypes.c_uint32),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# typedef struct _lck_attr_ {
#     unsigned int    lck_attr_val;
# } lck_attr_t;

class lck_attr_t(ctypes.Structure):
    _fields_ = (
        ("lck_attr_val", ctypes.c_uint32),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# typedef struct _lck_mtx_ {
#     union {
#         struct {
#             volatile uintptr_t      lck_mtx_owner;
#             union {
#                 struct {
#                     volatile uint32_t
#                         lck_mtx_waiters:16,
#                         lck_mtx_pri:8,
#                         lck_mtx_ilocked:1,
#                         lck_mtx_mlocked:1,
#                         lck_mtx_promoted:1,
#                         lck_mtx_spin:1,
#                         lck_mtx_is_ext:1,
#                         lck_mtx_pad3:3;
#                 };
#                     uint32_t    lck_mtx_state;
#             };
#             /* Pad field used as a canary, initialized to ~0 */
#             uint32_t            lck_mtx_pad32;
#         };
#         struct {
#             struct _lck_mtx_ext_        *lck_mtx_ptr;
#             uint32_t            lck_mtx_tag;
#             uint32_t            lck_mtx_pad32_2;
#         };
#     };
# } lck_mtx_t;

class lck_mtx_t(ctypes.Structure):
    class lck_mtx_tmp_union(ctypes.Union):
        class lck_mtx_tmp_union_struct_1(ctypes.Structure):
            class lck_mtx_tmp_union_struct_union(ctypes.Union):
                class lck_mtx_tmp_union_struct_union_struct(ctypes.Structure):
                    _fields_ = [
                        ("lck_mtx_waiters", ctypes.c_uint32, 16),
                        ("lck_mtx_pri", ctypes.c_uint32, 8),
                        ("lck_mtx_ilocked", ctypes.c_uint32, 1),
                        ("lck_mtx_mlocked", ctypes.c_uint32, 1),
                        ("lck_mtx_promoted", ctypes.c_uint32, 1),
                        ("lck_mtx_spin", ctypes.c_uint32, 1),
                        ("lck_mtx_is_ext", ctypes.c_uint32, 1),
                        ("lck_mtx_pad3", ctypes.c_uint32, 3)
                    ]
                _anonymous_ = ("tmp_struct", )
                _fields_ = (
                    ("tmp_struct", lck_mtx_tmp_union_struct_union_struct),
                    ("lck_mtx_state", ctypes.c_uint32)
                )
            _anonymous_ = ("tmp_union", )
            _fields_ = (
                ("lck_mtx_owner", ctypes.c_uint64),
                ("tmp_union", lck_mtx_tmp_union_struct_union),
                ("lck_mtx_pad32", ctypes.c_uint32)
            )

        class lck_mtx_tmp_union_struct_2(ctypes.Structure):
            _fields_ = (
                ("lck_mtx_ptr", POINTER64),
                ("lck_mtx_tag", ctypes.c_uint32),
                ("lck_mtx_pad32_2", ctypes.c_uint32)
            )

        _anonymous_ = ("tmp_struct_1", "tmp_struct_2", )
        _fields_ = (
            ("tmp_struct_1", lck_mtx_tmp_union_struct_1),
            ("tmp_struct_2", lck_mtx_tmp_union_struct_2),
        )
    _anonymous_ = ("tmp_union", )
    _fields_ = (
        ("tmp_union", lck_mtx_tmp_union),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# typedef struct {
#     unsigned int        type;
#     unsigned int        pad4;
#     vm_offset_t     pc;
#     vm_offset_t     thread;
# } lck_mtx_deb_t;

class lck_mtx_deb_t(ctypes.Structure):
    _fields_ = (
        ("type", ctypes.c_uint32),
        ("pad4", ctypes.c_uint32),
        ("pc", ctypes.c_uint64),
        ("thread", ctypes.c_uint64)
    )

# typedef struct _lck_mtx_ext_ {
#     lck_mtx_t       lck_mtx;
#     struct _lck_grp_    *lck_mtx_grp;
#     unsigned int        lck_mtx_attr;
#     unsigned int        lck_mtx_pad1;
#     lck_mtx_deb_t       lck_mtx_deb;
#     uint64_t        lck_mtx_stat;
#     unsigned int        lck_mtx_pad2[2];
# } lck_mtx_ext_t;

class lck_mtx_ext_t(ctypes.Structure):
    _fields_ = (
        ("lck_mtx", lck_mtx_t),
        ("lck_mtx_grp", POINTER64),
        ("lck_mtx_attr", ctypes.c_uint32),
        ("lck_mtx_pad1", ctypes.c_uint32),
        ("lck_mtx_deb", lck_mtx_deb_t),
        ("lck_mtx_stat", ctypes.c_uint64),
        ("lck_mtx_pad2", ctypes.c_uint32 * 2)
    )
    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct kauth_listener {
#     TAILQ_ENTRY(kauth_listener) kl_link;
#     const char *                kl_identifier;
#     kauth_scope_callback_t      kl_callback;
#     void *                      kl_idata;
# };

class kauth_listener_t(ctypes.Structure):
    class kauth_listener_struct(ctypes.Structure):
        _fields_ = (
            ("tqe_next", POINTER64),
            ("tqe_prev", POINTER64),
        )
    _fields_ = (
        ("kl_link", kauth_listener_struct),
        ("kl_identifier", POINTER64),
        ("kl_callback", POINTER64),
        ("kl_idata", POINTER64)
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct kauth_local_listener {
#     kauth_listener_t            kll_listenerp;
#     kauth_scope_callback_t      kll_callback;
#     void *                      kll_idata;
# };

class kauth_local_listener_t(ctypes.Structure):
    _fields_ = (
        ("kll_listenerp", kauth_listener_t),
        ("kll_callback", POINTER64),
        ("kll_idata", POINTER64),
    )

# struct kauth_scope {
#     TAILQ_ENTRY(kauth_scope)    ks_link;
#     volatile struct kauth_local_listener  ks_listeners[KAUTH_SCOPE_MAX_LISTENERS];
#     const char *                ks_identifier;
#     kauth_scope_callback_t      ks_callback;
#     void *                      ks_idata;
#     u_int                       ks_flags;
# };

class kauth_scope(ctypes.Structure):
    class kauth_scope_struct(ctypes.Structure):
        _fields_ = (
            ("tqe_next", POINTER64),
            ("tqe_prev", POINTER64),
        )
    _fields_ = (
        ("ks_link", kauth_scope_struct),
        ("ks_listeners", kauth_local_listener_t * 15),
        ("ks_identifier", POINTER64),
        ("ks_callback", POINTER64),
        ("ks_idata", POINTER64),
        ("ks_flags", ctypes.c_uint32)
    )
    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct domain {
#     int dom_family;     /* AF_xxx */
#     uint32_t dom_flags;     /* domain flags (see below ) */
#     uint32_t dom_refs;      /* # socreates outstanding */
#     lck_mtx_t *dom_mtx;     /* domain global mutex */
#     decl_lck_mtx_data(, dom_mtx_s);
#     TAILQ_ENTRY(domain) dom_entry;  /* next domain in list */
#     TAILQ_HEAD(, protosw) dom_protosw; /* protosw chain */
#     void    (*dom_init)     /* initialize domain data structures */
#             (struct domain *);
#     int (*dom_externalize)  /* externalize access rights */
#             (struct mbuf *);
#     void    (*dom_dispose)      /* dispose of internalized rights */
#             (struct mbuf *);
#     int (*dom_rtattach)     /* initialize routing table */
#             (void **, int);
#     int dom_rtoffset;       /* an arg to rtattach, in bits */
#     int dom_maxrtkey;       /* for routing layer */
#     int dom_protohdrlen;    /* len of protocol header */
#     const char *dom_name;
#     struct domain_old *dom_old; /* domain pointer per net_add_domain */
# };

class domain_t(ctypes.Structure):
    class domain_struct(ctypes.Structure):
        _fields_ = (
            ("tqe_next", POINTER64),
            ("tqe_prev", POINTER64),
        )

    _fields_ = (
        ("dom_family", ctypes.c_int32),
        ("dom_flags", ctypes.c_uint32),
        ("dom_refs", ctypes.c_uint32),
        ("dom_mtx", POINTER64),
        ("dom_mtx_s", lck_mtx_t),
        ("dom_entry", domain_struct),
        ("dom_protosw", tailq_head),
        ("dom_init", POINTER64),
        ("dom_externalize", POINTER64),
        ("dom_dispose", POINTER64),
        ("dom_rtattach", POINTER64),
        ("dom_rtoffset", ctypes.c_int32),
        ("dom_maxrtkey", ctypes.c_int32),
        ("dom_protohdrlen", ctypes.c_int32),
        ("dom_name", POINTER64),
        ("dom_old", POINTER64)
    )
    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

    def dump(self):
        for field in self._fields_:
            print(field[0], getattr(self, field[0]))

# struct kev_d_vectors {
#     u_int32_t   data_length;    /* Length of the event data */
#     void        *data_ptr;  /* Pointer to event data */
# };

# /*!
#     @struct kev_msg
#     @discussion This structure is used when posting a kernel event.
#     @field vendor_code The vendor code assigned by kev_vendor_code_find.
#     @field kev_class The event's class.
#     @field kev_class The event's subclass.
#     @field kev_class The event's code.
#     @field dv An array of vectors describing additional data to be appended
#         to the kernel event.
#  */
# struct kev_msg {
#     u_int32_t vendor_code;      /* For non-Apple extensibility */
#     u_int32_t kev_class;        /* Layer of event source */
#     u_int32_t kev_subclass;     /* Component within layer */
#     u_int32_t event_code;       /* The event code */
#     struct kev_d_vectors dv[N_KEV_VECTORS]; /* Up to n data vectors */
# };

class kev_d_vectors(ctypes.Structure):
    _fields_ = (
        ("data_length", ctypes.c_uint32),
        ("data_ptr", POINTER64),
    )

class kev_msg(ctypes.Structure):
    _fields_ = (
        ("vendor_code", ctypes.c_uint32),
        ("kev_class", ctypes.c_uint32),
        ("kev_subclass", ctypes.c_uint32),
        ("event_code", ctypes.c_uint32),
        ("dv", kev_d_vectors * 5)
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

    def dump(self):
        for field in self._fields_:
            print(field[0], getattr(self, field[0]))

# struct nlist_64 {
#     union {
#         uint32_t  n_strx; /* index into the string table */
#     } n_un;
#     uint8_t n_type;        /* type flag, see below */
#     uint8_t n_sect;        /* section number or NO_SECT */
#     uint16_t n_desc;       /* see <mach-o/stab.h> */
#     uint64_t n_value;      /* value of this symbol (or stab offset) */
# };

class nlist64_t(ctypes.Structure):
    _fields_ = (
        ("n_strx", ctypes.c_uint32),
        ("n_type", ctypes.c_uint8),
        ("n_sect", ctypes.c_uint8),
        ("n_desc", ctypes.c_uint16),
        ("n_value", ctypes.c_uint64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

    def __str__(self):
        return "\tn_strx = %d" % self.n_strx + "\n\tn_type = %d" % self.n_type + "\n\tn_sect = %d" % self.n_sect + "\n\tn_desc = %d" % self.n_desc + "\n\tn_value = 0x%x" % self.n_value

# struct sysent {        /* system call table */
#     sy_call_t    *sy_call;    /* implementing function */
#     sy_munge_t    *sy_arg_munge32; /* system call arguments munger for 32-bit process */
#     int32_t        sy_return_type; /* system call return types */
#     int16_t        sy_narg;    /* number of args */
#     uint16_t    sy_arg_bytes;    /* Total size of arguments in bytes for
#                      * 32-bit system calls
#                      */
# };

class sysent_t(ctypes.Structure):
    _fields_ = (
        ("sy_call", POINTER64),
        ("sy_arg_munge32", POINTER64),
        ("sy_return_type", ctypes.c_int32),
        ("sy_narg", ctypes.c_int16),
        ("sy_arg_bytes", ctypes.c_uint16),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct proc {
#     LIST_ENTRY(proc) p_list;        /* List of all processes. */

#     pid_t        p_pid;            /* Process identifier. (static)*/
#     void *         task;            /* corresponding task (static)*/
#     struct    proc *    p_pptr;             /* Pointer to parent process.(LL) */
#     pid_t        p_ppid;            /* process's parent pid number */
#     pid_t        p_pgrpid;        /* process group id of the process (LL)*/
#     uid_t        p_uid;
#     gid_t        p_gid;
#     uid_t        p_ruid;
#     gid_t        p_rgid;
#     uid_t        p_svuid;
#     gid_t        p_svgid;
#     uint64_t    p_uniqueid;        /* process unique ID - incremented on fork/spawn/vfork, remains same across exec. */
#     uint64_t    p_puniqueid;        /* parent's unique ID - set on fork/spawn/vfork, doesn't change if reparented. */

#     struct lck_mtx_t     p_mlock;        /* mutex lock for proc */

#     char        p_stat;            /* S* process status. (PL)*/
#     char        p_shutdownstate;
#     char        p_kdebug;        /* P_KDEBUG eq (CC)*/
#     char        p_btrace;        /* P_BTRACE eq (CC)*/

#     LIST_ENTRY(proc) p_pglist;        /* List of processes in pgrp.(PGL) */
#     LIST_ENTRY(proc) p_sibling;        /* List of sibling processes. (LL)*/
#     LIST_HEAD(, proc) p_children;        /* Pointer to list of children. (LL)*/
#     TAILQ_HEAD( , uthread) p_uthlist;     /* List of uthreads  (PL) */

#     LIST_ENTRY(proc) p_hash;        /* Hash chain. (LL)*/
#     TAILQ_HEAD( ,eventqelt) p_evlist;    /* (PL) */

#     struct lck_mtx_t    p_fdmlock;        /* proc lock to protect fdesc */
#     struct lck_mtx_t     p_ucred_mlock;        /* mutex lock to protect p_ucred */

#     /* substructures: */
#     kauth_cred_t    p_ucred;        /* Process owner's identity. (PUCL) */
#     struct    filedesc *p_fd;            /* Ptr to open files structure. (PFDL) */
#     struct    pstats *p_stats;        /* Accounting/statistics (PL). */
#     struct    plimit *p_limit;        /* Process limits.(PL) */

#     struct    sigacts *p_sigacts;        /* Signal actions, state (PL) */
#      int        p_siglist;        /* signals captured back from threads */
#     struct lck_spin_t    p_slock;        /* spin lock for itimer/profil protection */

# #define    p_rlimit    p_limit->pl_rlimit

#     struct    plimit *p_olimit;        /* old process limits  - not inherited by child  (PL) */
#     unsigned int    p_flag;            /* P_* flags. (atomic bit ops) */
#     unsigned int    p_lflag;        /* local flags  (PL) */
#     unsigned int    p_listflag;        /* list flags (LL) */
#     unsigned int    p_ladvflag;        /* local adv flags (atomic) */
#     int        p_refcount;        /* number of outstanding users(LL) */
#     int        p_childrencnt;        /* children holding ref on parent (LL) */
#     int        p_parentref;        /* children lookup ref on parent (LL) */

#     pid_t        p_oppid;         /* Save parent pid during ptrace. XXX */
#     u_int        p_xstat;        /* Exit status for wait; also stop signal. */
#     uint8_t p_xhighbits;        /* Stores the top byte of exit status to avoid truncation*/

#     struct    itimerval p_realtimer;        /* Alarm timer. (PSL) */
#     struct    timeval p_rtime;        /* Real time.(PSL)  */
#     struct    itimerval p_vtimer_user;    /* Virtual timers.(PSL)  */
#     struct    itimerval p_vtimer_prof;    /* (PSL) */

#     struct    timeval    p_rlim_cpu;        /* Remaining rlim cpu value.(PSL) */
#     int        p_debugger;        /*  NU 1: can exec set-bit programs if suser */
#     boolean_t    sigwait;    /* indication to suspend (PL) */
#     void    *sigwait_thread;    /* 'thread' holding sigwait(PL)  */
#     void    *exit_thread;        /* Which thread is exiting(PL)  */
#     int    p_vforkcnt;        /* number of outstanding vforks(PL)  */
#         void *  p_vforkact;         /* activation running this vfork proc)(static)  */
#     int    p_fpdrainwait;        /* (PFDL) */
#     pid_t    p_contproc;    /* last PID to send us a SIGCONT (PL) */

#     /* Following fields are info from SIGCHLD (PL) */
#     pid_t    si_pid;            /* (PL) */
#     u_int   si_status;        /* (PL) */
#     u_int    si_code;        /* (PL) */
#     uid_t    si_uid;            /* (PL) */

#     void * vm_shm;            /* (SYSV SHM Lock) for sysV shared memory */

#     user_addr_t            p_dtrace_argv;            /* (write once, read only after that) */
#     user_addr_t            p_dtrace_envp;            /* (write once, read only after that) */
#     lck_mtx_t            p_dtrace_sprlock;        /* sun proc lock emulation */
#     int                p_dtrace_probes;        /* (PL) are there probes for this proc? */
#     u_int                p_dtrace_count;            /* (sprlock) number of DTrace tracepoints */
#         uint8_t                         p_dtrace_stop;                  /* indicates a DTrace-desired stop */
#     struct dtrace_ptss_page*    p_dtrace_ptss_pages;        /* (sprlock) list of user ptss pages */
#     struct dtrace_ptss_page_entry*    p_dtrace_ptss_free_list;    /* (atomic) list of individual ptss entries */
#     struct dtrace_helpers*        p_dtrace_helpers;        /* (dtrace_lock) DTrace per-proc private */
#     struct dof_ioctl_data*        p_dtrace_lazy_dofs;        /* (sprlock) unloaded dof_helper_t's */

# /* XXXXXXXXXXXXX BCOPY'ed on fork XXXXXXXXXXXXXXXX */
# /* The following fields are all copied upon creation in fork. */
# #define    p_startcopy    p_argslen

#     u_int    p_argslen;     /* Length of process arguments. */
#     int      p_argc;            /* saved argc for sysctl_procargs() */
#     user_addr_t user_stack;        /* where user stack was allocated */
#     struct    vnode *p_textvp;    /* Vnode of executable. */
#     off_t    p_textoff;        /* offset in executable vnode */

#     sigset_t p_sigmask;        /* DEPRECATED */
#     sigset_t p_sigignore;    /* Signals being ignored. (PL) */
#     sigset_t p_sigcatch;    /* Signals being caught by user.(PL)  */

#     u_char    p_priority;    /* (NU) Process priority. */
#     u_char    p_resv0;    /* (NU) User-priority based on p_cpu and p_nice. */
#     char    p_nice;        /* Process "nice" value.(PL) */
#     u_char    p_resv1;    /* (NU) User-priority based on p_cpu and p_nice. */

#     // types currently in sys/param.h
#     command_t   p_comm;
#     proc_name_t p_name;    /* can be changed by the process */


#     struct     pgrp *p_pgrp;        /* Pointer to process group. (LL) */
#     uint32_t    p_csflags;    /* flags for codesign (PL) */
#     uint32_t    p_pcaction;    /* action  for process control on starvation */
#     uint8_t p_uuid[16];        /* from LC_UUID load command */

#     /*
#      * CPU type and subtype of binary slice executed in
#      * this process.  Protected by proc lock.
#      */
#     cpu_type_t    p_cputype;
#     cpu_subtype_t    p_cpusubtype;

# /* End area that is copied on creation. */
# /* XXXXXXXXXXXXX End of BCOPY'ed on fork (AIOLOCK)XXXXXXXXXXXXXXXX */
# #define    p_endcopy    p_aio_total_count
#     int        p_aio_total_count;        /* all allocated AIO requests for this proc */
#     int        p_aio_active_count;        /* all unfinished AIO requests for this proc */
#     TAILQ_HEAD( , aio_workq_entry ) p_aio_activeq;     /* active async IO requests */
#     TAILQ_HEAD( , aio_workq_entry ) p_aio_doneq;    /* completed async IO requests */

#     SLIST_HEAD(klist, klist) p_klist;  /* knote list (PL ?)*/

#     struct    rusage_superset *p_ru;    /* Exit information. (PL) */
#     int        p_sigwaitcnt;
#     thread_t     p_signalholder;
#     thread_t     p_transholder;

#     /* DEPRECATE following field  */
#     u_short    p_acflag;    /* Accounting flags. */
#     volatile u_short p_vfs_iopolicy;    /* VFS iopolicy flags. (atomic bit ops) */

#     user_addr_t     p_threadstart;        /* pthread start fn */
#     user_addr_t     p_wqthread;        /* pthread workqueue fn */
#     int     p_pthsize;            /* pthread size */
#     uint32_t    p_pth_tsd_offset;    /* offset from pthread_t to TSD for new threads */
#     user_addr_t    p_stack_addr_hint;    /* stack allocation hint for wq threads */
#     void *     p_wqptr;            /* workq ptr */

#     struct  timeval p_start;            /* starting time */
#     void *    p_rcall;
#     int        p_ractive;
#     int    p_idversion;        /* version of process identity */
#     void *    p_pthhash;            /* pthread waitqueue hash */
#     volatile uint64_t was_throttled __attribute__((aligned(8))); /* Counter for number of throttled I/Os */
#     volatile uint64_t did_throttle __attribute__((aligned(8)));  /* Counter for number of I/Os this proc throttled */

#     uint64_t    p_dispatchqueue_offset;
#     uint64_t    p_dispatchqueue_serialno_offset;
#     uint64_t    p_return_to_kernel_offset;
#     uint64_t    p_mach_thread_self_offset;

#     struct timeval    vm_pressure_last_notify_tstamp;

#     /* Fields protected by proc list lock */
#     TAILQ_ENTRY(proc) p_memstat_list;               /* priority bucket link */
#     uint32_t          p_memstat_state;              /* state */
#     int32_t           p_memstat_effectivepriority;  /* priority after transaction state accounted for */
#     int32_t           p_memstat_requestedpriority;  /* active priority */
#     uint32_t          p_memstat_dirty;              /* dirty state */
#     uint64_t          p_memstat_userdata;           /* user state */
#     uint64_t          p_memstat_idledeadline;       /* time at which process became clean */
#     uint64_t          p_memstat_idle_start;         /* abstime process transitions into the idle band */
#     uint64_t      p_memstat_idle_delta;         /* abstime delta spent in idle band */
#     int32_t           p_memstat_memlimit;           /* cached memory limit, toggles between active and inactive limits */
#     int32_t           p_memstat_memlimit_active;    /* memory limit enforced when process is in active jetsam state */
#     int32_t           p_memstat_memlimit_inactive;    /* memory limit enforced when process is in inactive jetsam state */

#     /* cached proc-specific data required for corpse inspection */
#     pid_t             p_responsible_pid;    /* pid resonsible for this process */
#     _Atomic uint32_t  p_user_faults; /* count the number of user faults generated */

#     struct os_reason     *p_exit_reason;

#     uint64_t    p_user_data;            /* general-purpose storage for userland-provided data */

# };

class timeval_t(ctypes.Structure):
    _fields_ = (
        ("tv_sec", ctypes.c_long),
        ("tv_usec", ctypes.c_int32),
    )

class itimerval_t(ctypes.Structure):
    _fields_ = (
        ("it_interval", timeval_t),
        ("it_value", timeval_t),
    )

class proc_t(ctypes.Structure):
    # struct lck_spin_t {
    #     unsigned long    opaque[10];
    # };
    class lck_spin_t(ctypes.Structure):
        _fields_ = (
            ("opaque", ctypes.c_ulong * 10),
        )
    _fields_ = (
        ("p_list", list_entry),
        ("p_pid", ctypes.c_int32),
        ("task", POINTER64),
        ("p_pptr", POINTER64),
        ("p_ppid", ctypes.c_int32),
        ("p_pgrpid", ctypes.c_int32),
        ("p_uid", ctypes.c_uint32),
        ("p_gid", ctypes.c_uint32),
        ("p_ruid", ctypes.c_uint32),
        ("p_rgid", ctypes.c_uint32),
        ("p_svuid", ctypes.c_uint32),
        ("p_svgid", ctypes.c_uint32),
        ("p_uniqueid", ctypes.c_uint64),
        ("p_puniqueid", ctypes.c_uint64),
        ("p_mlock", lck_mtx_t),
        ("p_stat", ctypes.c_char),
        ("p_shutdownstate", ctypes.c_char),
        ("p_kdebug", ctypes.c_char),
        ("p_btrace", ctypes.c_char),
        ("p_pglist", list_entry),
        ("p_sibling", list_entry),
        ("p_children", list_head),
        ("p_uthlist", tailq_head),
        ("p_hash", list_entry),
        ("p_evlist", tailq_head),
        ("p_fdmlock", lck_mtx_t),
        ("p_ucred_mlock", lck_mtx_t),
        ("p_ucred", POINTER64),
        ("p_fd", POINTER64),
        ("p_stats", POINTER64),
        ("p_limit", POINTER64),
        ("p_sigacts", POINTER64),
        ("p_siglist", ctypes.c_int),
        ("p_slock", lck_spin_t),
        ("p_olimit", POINTER64),
        ("p_flag", ctypes.c_uint),
        ("p_lflag", ctypes.c_uint),
        ("p_listflag", ctypes.c_uint),
        ("p_ladvflag", ctypes.c_uint),
        ("p_refcount", ctypes.c_int),
        ("p_childrencnt", ctypes.c_int),
        ("p_parentref", ctypes.c_int),
        ("p_oppid", ctypes.c_int32),
        ("p_xstat", ctypes.c_uint),
        ("p_xhighbits", ctypes.c_uint8),
        ("p_realtimer", itimerval_t),
        ("p_rtime", timeval_t),
        ("p_vtimer_user", itimerval_t),
        ("p_vtimer_prof", itimerval_t),
        ("p_rlim_cpu", timeval_t),
        ("p_debugger", ctypes.c_int),
        ("sigwait", ctypes.c_int),
        ("sigwait_thread", POINTER64),
        ("exit_thread", POINTER64),
        ("p_vforkcnt", ctypes.c_int),
        ("p_vforkact", POINTER64),
        ("p_fpdrainwait", ctypes.c_int),
        ("p_contproc", ctypes.c_int32),
        ("si_pid", ctypes.c_int32),
        ("si_status", ctypes.c_uint),
        ("si_code", ctypes.c_uint),
        ("si_uid", ctypes.c_uint32),
        ("vm_shm", POINTER64),
        ("p_dtrace_argv", ctypes.c_uint64),
        ("p_dtrace_envp", ctypes.c_uint64),
        ("p_dtrace_sprlock", lck_mtx_t),
        ("p_dtrace_probes", ctypes.c_int),
        ("p_dtrace_count", ctypes.c_uint),
        ("p_dtrace_stop", ctypes.c_uint8),
        ("p_dtrace_ptss_pages", POINTER64),
        ("p_dtrace_ptss_free_list", POINTER64),
        ("p_dtrace_helpers", POINTER64),
        ("p_dtrace_lazy_dofs", POINTER64),
        ("p_argslen", ctypes.c_uint),
        ("p_argc", ctypes.c_int),
        ("user_stack", ctypes.c_uint64),
        ("p_textvp", POINTER64),
        ("p_textoff", ctypes.c_int64),
        ("p_sigmask", ctypes.c_uint32),
        ("p_sigignore", ctypes.c_uint32),
        ("p_sigcatch", ctypes.c_uint32),
        ("p_priority", ctypes.c_ubyte),
        ("p_resv0", ctypes.c_ubyte),
        ("p_nice", ctypes.c_char),
        ("p_resv1", ctypes.c_ubyte),
        ("p_comm", ctypes.c_char * (16 + 1)),
        ("p_name", ctypes.c_char * (2*16 + 1)),
        ("p_pgrp", POINTER64),
        ("p_csflags", ctypes.c_uint32),
        ("p_pcaction", ctypes.c_uint32),
        ("p_uuid", ctypes.c_uint8 * 16),
        ("p_cputype", ctypes.c_int),
        ("p_cpusubtype", ctypes.c_int),
        ("p_aio_total_count", ctypes.c_int),
        ("p_aio_active_count", ctypes.c_int),
        ("p_aio_activeq", tailq_head),
        ("p_aio_doneq", tailq_head),
        ("p_klist", slist_head),
        ("p_ru", POINTER64),
        ("p_sigwaitcnt", ctypes.c_int),
        ("p_signalholder", POINTER64),
        ("p_transholder", POINTER64),
        ("p_acflag", ctypes.c_ushort),
        ("p_vfs_iopolicy", ctypes.c_ushort),
        ("p_threadstart", ctypes.c_uint64),
        ("p_wqthread", ctypes.c_uint64),
        ("p_pthsize", ctypes.c_int),
        ("p_pth_tsd_offset", ctypes.c_uint32),
        ("p_stack_addr_hint", ctypes.c_uint64),
        ("p_wqptr", POINTER64),
        ("p_start", timeval_t),
        ("p_rcall", POINTER64),
        ("p_ractive", ctypes.c_int),
        ("p_idversion", ctypes.c_int),
        ("p_pthhash", POINTER64),
        ("was_throttled", ctypes.c_uint64),
        ("did_throttle", ctypes.c_uint64),
        ("p_dispatchqueue_offset", ctypes.c_uint64),
        ("p_dispatchqueue_serialno_offset", ctypes.c_uint64),
        ("p_return_to_kernel_offset", ctypes.c_uint64),
        ("p_mach_thread_self_offset", ctypes.c_uint64),
        ("vm_pressure_last_notify_tstamp", timeval_t),
        ("p_memstat_list", tailq_entry),
        ("p_memstat_state", ctypes.c_uint32),
        ("p_memstat_effectivepriority", ctypes.c_int32),
        ("p_memstat_requestedpriority", ctypes.c_int32),
        ("p_memstat_dirty", ctypes.c_uint32),
        ("p_memstat_userdata", ctypes.c_uint64),
        ("p_memstat_idledeadline", ctypes.c_uint64),
        ("p_memstat_idle_start", ctypes.c_uint64),
        ("p_memstat_idle_delta", ctypes.c_uint64),
        ("p_memstat_memlimit", ctypes.c_int32),
        ("p_memstat_memlimit_active", ctypes.c_int32),
        ("p_memstat_memlimit_inactive", ctypes.c_int32),
        ("p_responsible_pid", ctypes.c_int32),
        ("p_user_faults", ctypes.c_uint32),
        ("p_exit_reason", POINTER64),
        ("p_user_data", ctypes.c_uint64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# _STRUCT_TIMESPEC
# {
#     __darwin_time_t tv_sec;
#     long            tv_nsec;
# };
class timespec_t(ctypes.Structure):
    _fields_ = (
        ("tv_sec", ctypes.c_long),
        ("tv_nsec", ctypes.c_long),
    )

# typedef struct __attribute__((packed)) {
#     uint32_t length;

#     /* common attributes */
#     attribute_set_t attrset;
#     attrreference_t name;
#     dev_t st_dev;
#     fsobj_type_t objtype;
#     struct timespec st_birthtimespec;
#     struct timespec st_mtimespec;
#     struct timespec st_ctimespec;
#     struct timespec st_atimespec;
#     uid_t st_uid;
#     gid_t st_gid;
#     uint32_t accessmask;
#     uint32_t st_flags;
#     uint64_t st_ino;

#     /* non-directory attributes */
#     uint32_t st_nlink;
#     off_t allocsize;
#     uint32_t st_blksize;
#     uint32_t st_rdev;
#     off_t st_size;
# } attrListAttributes;

class attrListAttributes_t(ctypes.Structure):
    # typedef struct attribute_set {
    #     attrgroup_t commonattr;         /* common attribute group */
    #     attrgroup_t volattr;            /* Volume attribute group */
    #     attrgroup_t dirattr;            /* directory attribute group */
    #     attrgroup_t fileattr;           /* file attribute group */
    #     attrgroup_t forkattr;           /* fork attribute group */
    # } attribute_set_t;
    class attribute_set_t(ctypes.Structure):
        _fields_ = (
            ("commonattr", ctypes.c_uint32),
            ("volattr", ctypes.c_uint32),
            ("dirattr", ctypes.c_uint32),
            ("fileattr", ctypes.c_uint32),
            ("forkattr", ctypes.c_uint32),
        )

    # typedef struct attrreference {
    #     int32_t     attr_dataoffset;
    #     u_int32_t   attr_length;
    # } attrreference_t;
    class attrreference_t(ctypes.Structure):
        _fields_ = (
            ("attr_dataoffset", ctypes.c_int32),
            ("attr_length", ctypes.c_uint32),
        )

    _fields_ = (
        ("length", ctypes.c_uint32),
        ("attrset", attribute_set_t),
        ("name", attrreference_t),
        ("st_dev", ctypes.c_int32),
        ("objtype", ctypes.c_uint32),
        ("st_birthtimespec", timespec_t),
        ("st_mtimespec", timespec_t),
        ("st_ctimespec", timespec_t),
        ("st_atimespec", timespec_t),
        ("st_uid", ctypes.c_uint32),
        ("st_gid", ctypes.c_uint32),
        ("accessmask", ctypes.c_uint32),
        ("st_flags", ctypes.c_uint32),
        ("st_ino", ctypes.c_uint64),
        ("st_nlink", ctypes.c_uint32),
        ("allocsize", ctypes.c_int64),
        ("st_blksize", ctypes.c_uint32),
        ("st_rdev", ctypes.c_uint32),
        ("st_size", ctypes.c_uint32)
    )

# struct attrlist {
#     u_short bitmapcount;            /* number of attr. bit sets in list (should be 5) */
#     u_int16_t reserved;         /* (to maintain 4-byte alignment) */
#     attrgroup_t commonattr;         /* common attribute group */
#     attrgroup_t volattr;            /* Volume attribute group */
#     attrgroup_t dirattr;            /* directory attribute group */
#     attrgroup_t fileattr;           /* file attribute group */
#     attrgroup_t forkattr;           /* fork attribute group */
# };

class attrlist_t(ctypes.Structure):
    _fields_ = (
        ("bitmapcount", ctypes.c_ushort),
        ("reserved", ctypes.c_uint16),
        ("commonattr", ctypes.c_uint32),
        ("volattr", ctypes.c_uint32),
        ("dirattr", ctypes.c_uint32),
        ("fileattr", ctypes.c_uint32),
        ("forkattr", ctypes.c_uint32),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct getattrlistbulk_args {
#     char dirfd_l_[PADL_(int)]; int dirfd; char dirfd_r_[PADR_(int)];
#     char alist_l_[PADL_(user_addr_t)]; user_addr_t alist; char alist_r_[PADR_(user_addr_t)];
#     char attributeBuffer_l_[PADL_(user_addr_t)]; user_addr_t attributeBuffer; char attributeBuffer_r_[PADR_(user_addr_t)];
#     char bufferSize_l_[PADL_(user_size_t)]; user_size_t bufferSize; char bufferSize_r_[PADR_(user_size_t)];
#     char options_l_[PADL_(uint64_t)]; uint64_t options; char options_r_[PADR_(uint64_t)];
# };

class getattrlistbulk_args_t(ctypes.Structure):
    _fields_ = (
        ("dirfd", ctypes.c_int),
        ("dirfd_r_", ctypes.c_byte * 4),
        ("alist", ctypes.c_uint64),
        ("attributeBuffer", ctypes.c_uint64),
        ("bufferSize", ctypes.c_uint64),
        ("options", ctypes.c_uint64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# /* use kern_iovec for system space requests */
# struct kern_iovec {
#     u_int64_t   iov_base;   /* Base address. */
#     u_int64_t   iov_len;    /* Length. */
# };

# /* use user_iovec for user space requests */
# struct user_iovec {
#     user_addr_t iov_base;   /* Base address. */
#     user_size_t iov_len;    /* Length. */
# };

# union iovecs {
#     struct kern_iovec   *kiovp;
#     struct user_iovec   *uiovp;
# };

# /* WARNING - use accessor calls for uio_iov and uio_resid since these */
# /* fields vary depending on the originating address space. */
# struct uio {
#     union iovecs    uio_iovs;       /* current iovec */
#     int             uio_iovcnt;     /* active iovecs */
#     off_t           uio_offset;
#     enum uio_seg    uio_segflg;
#     enum uio_rw     uio_rw;
#     user_size_t uio_resid_64;
#     int             uio_size;       /* size for use with kfree */
#     int             uio_max_iovs;   /* max number of iovecs this uio_t can hold */
#     u_int32_t       uio_flags;      
# };

class kern_iovec_t(ctypes.Structure):
    _fields_ = (
        ("iov_base", ctypes.c_uint64),
        ("iov_len", ctypes.c_uint64),
    )

class user_iovec_t(ctypes.Structure):
    _fields_ = (
        ("iov_base", ctypes.c_uint64),
        ("iov_len", ctypes.c_uint64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

class iovecs_t(ctypes.Union):
    _fields_ = (
        ("kiovp", POINTER64),
        ("uiovp", POINTER64),
    )

class uio_t(ctypes.Structure):
    _fields_ = (
        ("uio_iovs", iovecs_t),
        ("uio_iovcnt", ctypes.c_int),
        ("uio_offset", ctypes.c_int64),
        ("uio_segflg", ctypes.c_int),
        ("uio_rw", ctypes.c_int),
        ("uio_resid_64", ctypes.c_uint64),
        ("uio_size", ctypes.c_int),
        ("uio_max_iovs", ctypes.c_int),
        ("uio_flags", ctypes.c_uint32),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj

# struct vnode_attr {
#     uint64_t    va_supported;
#     uint64_t    va_active;
#     int     va_vaflags;
    
#     dev_t       va_rdev;    /* device id (device nodes only) */
#     uint64_t    va_nlink;   /* number of references to this file */
#     uint64_t    va_total_size;  /* size in bytes of all forks */
#     uint64_t    va_total_alloc; /* disk space used by all forks */
#     uint64_t    va_data_size;   /* size in bytes of the fork managed by current vnode */
#     uint64_t    va_data_alloc;  /* disk space used by the fork managed by current vnode */
#     uint32_t    va_iosize;  /* optimal I/O blocksize */

#     uid_t       va_uid;     /* owner UID */
#     gid_t       va_gid;     /* owner GID */
#     mode_t      va_mode;    /* posix permissions */
#     uint32_t    va_flags;   /* file flags */
#     struct kauth_acl *va_acl;   /* access control list */

#     struct timespec va_create_time; /* time of creation */
#     struct timespec va_access_time; /* time of last access */
#     struct timespec va_modify_time; /* time of last data modification */
#     struct timespec va_change_time; /* time of last metadata change */
#     struct timespec va_backup_time; /* time of last backup */
    
#     uint64_t    va_fileid;  /* file unique ID in filesystem */
#     uint64_t    va_linkid;  /* file link unique ID */
#     uint64_t    va_parentid;    /* parent ID */
#     uint32_t    va_fsid;    /* filesystem ID */
#     uint64_t    va_filerev; /* file revision counter */ /* XXX */
#     uint32_t    va_gen;     /* file generation count */ /* XXX - relationship of
#                                     * these two? */
#     uint32_t    va_encoding;    /* filename encoding script */

#     enum vtype  va_type;    /* file type */
#     char *      va_name;    /* Name for ATTR_CMN_NAME; MAXPATHLEN bytes */
#     guid_t      va_uuuid;   /* file owner UUID */
#     guid_t      va_guuid;   /* file group UUID */
    
#     uint64_t    va_nchildren;     /* Number of items in a directory */
#     uint64_t    va_dirlinkcount;  /* Real references to dir (i.e. excluding "." and ".." refs) */

#     struct kauth_acl *va_base_acl;
#     struct timespec va_addedtime;   /* timestamp when item was added to parent directory */
        
#     uint32_t va_dataprotect_class;  /* class specified for this file if it didn't exist */
#     uint32_t va_dataprotect_flags;  /* flags from NP open(2) to the filesystem */

#     uint32_t va_document_id;

#     uint32_t    va_devid;   /* devid of filesystem */
#     uint32_t    va_objtype; /* type of object */
#     uint32_t    va_objtag;  /* vnode tag of filesystem */
#     uint32_t    va_user_access; /* access for user */
#     uint8_t     va_finderinfo[32];  /* Finder Info */
#     uint64_t    va_rsrc_length; /* Resource Fork length */
#     uint64_t    va_rsrc_alloc;  /* Resource Fork allocation size */
#     fsid_t      va_fsid64;  /* fsid, of the correct type  */

#     uint32_t va_write_gencount;     /* counter that increments each time the file changes */

#     uint64_t va_private_size; /* If the file were deleted, how many bytes would be freed immediately */
# };

class vnode_attr_t(ctypes.Structure):
    # typedef struct {
    # #define KAUTH_GUID_SIZE 16  /* 128-bit identifier */
    #     unsigned char g_guid[KAUTH_GUID_SIZE];
    # } guid_t;
    class guid_t(ctypes.Structure):
        _fields_ = (
            ("g_guid", ctypes.c_ubyte * 16),
        )

    # typedef struct fsid { int32_t val[2]; } fsid_t; /* file system id type */
    class fsid_t(ctypes.Structure):
        _fields_ = (
            ("val", ctypes.c_int32 * 2),
        )

    _fields_ = (
        ("va_supported", ctypes.c_uint64),
        ("va_active", ctypes.c_uint64),
        ("va_vaflags", ctypes.c_int),

        ("va_rdev", ctypes.c_int32),
        ("va_nlink", ctypes.c_uint64),
        ("va_total_size", ctypes.c_uint64),
        ("va_total_alloc", ctypes.c_uint64),
        ("va_data_size", ctypes.c_uint64),
        ("va_data_alloc", ctypes.c_uint64),
        ("va_iosize", ctypes.c_uint32),

        ("va_uid", ctypes.c_uint32),
        ("va_gid", ctypes.c_uint32),
        ("va_mode", ctypes.c_uint16),
        ("va_flags", ctypes.c_uint32),
        ("va_acl", POINTER64),

        ("va_create_time", timespec_t),
        ("va_access_time", timespec_t),
        ("va_modify_time", timespec_t),
        ("va_change_time", timespec_t),
        ("va_backup_time", timespec_t),

        ("va_fileid", ctypes.c_uint64),
        ("va_linkid", ctypes.c_uint64),
        ("va_parentid", ctypes.c_uint64),
        ("va_fsid", ctypes.c_uint32),
        ("va_filerev", ctypes.c_uint64),
        ("va_gen", ctypes.c_uint32),

        ("va_encoding", ctypes.c_uint32),

        ("va_type", ctypes.c_int),
        ("va_name", POINTER64),
        ("va_uuuid", guid_t),
        ("va_guuid", guid_t),

        ("va_nchildren", ctypes.c_uint64),
        ("va_dirlinkcount", ctypes.c_uint64),

        ("va_base_acl", POINTER64),
        ("va_addedtime", timespec_t),

        ("va_dataprotect_class", ctypes.c_uint32),
        ("va_dataprotect_flags", ctypes.c_uint32),
        ("va_document_id", ctypes.c_uint32),

        ("va_devid", ctypes.c_uint32),
        ("va_objtype", ctypes.c_uint32),
        ("va_objtag", ctypes.c_uint32),
        ("va_user_access", ctypes.c_uint32),
        ("va_finderinfo", ctypes.c_uint8 * 32),
        ("va_rsrc_length", ctypes.c_uint64),
        ("va_rsrc_alloc", ctypes.c_uint64),
        ("va_fsid64", fsid_t),

        ("va_write_gencount", ctypes.c_uint32),

        ("va_private_size", ctypes.c_uint64),
    )

    def __init__(self, ql, base):
        self.ql = ql
        self.base = base

    def updateToMem(self):
        self.ql.mem.write(self.base, bytes(self))

    def loadFromMem(self):
        data = self.ql.mem.read(self.base, ctypes.sizeof(self))
        newObj = type(self).from_buffer(data)
        newObj.ql = self.ql
        newObj.base = self.base
        return newObj


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from .ProcessorBind import *

# @see: MdePkg\Include\Uefi\UefiMultiPhase.h
class EFI_TABLE_HEADER(STRUCT):
	_fields_ = [
		('Signature',	UINT64),
		('Revision',	UINT32),
		('HeaderSize',	UINT32),
		('CRC32',		UINT32),
		('Reserved',	UINT32)
	]

class EFI_RESET_TYPE(ENUM):
	_members_ = [
		'EfiResetCold',
		'EfiResetWarm'
		'EfiResetShutdown',
		'EfiResetPlatformSpecific',
	]

class EFI_MEMORY_TYPE(ENUM):
	_members_ = [
		'EfiReservedMemoryType',
		'EfiLoaderCode',
		'EfiLoaderData',
		'EfiBootServicesCode',
		'EfiBootServicesData',
		'EfiRuntimeServicesCode',
		'EfiRuntimeServicesData',
		'EfiConventionalMemory',
		'EfiUnusableMemory',
		'EfiACPIReclaimMemory',
		'EfiACPIMemoryNVS',
		'EfiMemoryMappedIO',
		'EfiMemoryMappedIOPortSpace',
		'EfiPalCode',
		'EfiPersistentMemory',
		'EfiMaxMemoryType'
	]

__all__ = [
	'EFI_TABLE_HEADER',
	'EFI_RESET_TYPE',
	'EFI_MEMORY_TYPE'
]


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

# Linux syscall numbers

NR_read = 0
NR_write = 1
NR_open = 2
NR_close = 3
NR_stat = 4
NR_fstat = 5
NR_lstat = 6
NR_poll = 7
NR_lseek = 8
NR_mmap = 9
NR_mprotect = 10
NR_munmap = 11
NR_brk = 12
NR_rt_sigaction = 13
NR_rt_sigprocmask = 14
NR_rt_sigreturn = 15
NR_ioctl = 16
NR_pread64 = 17
NR_pwrite64 = 18
NR_readv = 19
NR_writev = 20
NR_access = 21
NR_pipe = 22
NR_select = 23
NR_sched_yield = 24
NR_mremap = 25
NR_msync = 26
NR_mincore = 27
NR_madvise = 28
NR_shmget = 29
NR_shmat = 30
NR_shmctl = 31
NR_dup = 32
NR_dup2 = 33
NR_pause = 34
NR_nanosleep = 35
NR_getitimer = 36
NR_alarm = 37
NR_setitimer = 38
NR_getpid = 39
NR_sendfile = 40
NR_socket = 41
NR_connect = 42
NR_accept = 43
NR_sendto = 44
NR_recvfrom = 45
NR_sendmsg = 46
NR_recvmsg = 47
NR_shutdown = 48
NR_bind = 49
NR_listen = 50
NR_getsockname = 51
NR_getpeername = 52
NR_socketpair = 53
NR_setsockopt = 54
NR_getsockopt = 55
NR_clone = 56
NR_fork = 57
NR_vfork = 58
NR_execve = 59
NR_exit = 60
NR_wait4 = 61
NR_kill = 62
NR_uname = 63
NR_semget = 64
NR_semop = 65
NR_semctl = 66
NR_shmdt = 67
NR_msgget = 68
NR_msgsnd = 69
NR_msgrcv = 70
NR_msgctl = 71
NR_fcntl = 72
NR_flock = 73
NR_fsync = 74
NR_fdatasync = 75
NR_truncate = 76
NR_ftruncate = 77
NR_getdents = 78
NR_getcwd = 79
NR_chdir = 80
NR_fchdir = 81
NR_rename = 82
NR_mkdir = 83
NR_rmdir = 84
NR_creat = 85
NR_link = 86
NR_unlink = 87
NR_symlink = 88
NR_readlink = 89
NR_chmod = 90
NR_fchmod = 91
NR_chown = 92
NR_fchown = 93
NR_lchown = 94
NR_umask = 95
NR_gettimeofday = 96
NR_getrlimit = 97
NR_getrusage = 98
NR_sysinfo = 99
NR_times = 100
NR_ptrace = 101
NR_getuid = 102
NR_syslog = 103
NR_getgid = 104
NR_setuid = 105
NR_setgid = 106
NR_geteuid = 107
NR_getegid = 108
NR_setpgid = 109
NR_getppid = 110
NR_getpgrp = 111
NR_setsid = 112
NR_setreuid = 113
NR_setregid = 114
NR_getgroups = 115
NR_setgroups = 116
NR_setresuid = 117
NR_getresuid = 118
NR_setresgid = 119
NR_getresgid = 120
NR_getpgid = 121
NR_setfsuid = 122
NR_setfsgid = 123
NR_getsid = 124
NR_capget = 125
NR_capset = 126
NR_rt_sigpending = 127
NR_rt_sigtimedwait = 128
NR_rt_sigqueueinfo = 129
NR_rt_sigsuspend = 130
NR_sigaltstack = 131
NR_utime = 132
NR_mknod = 133
NR_uselib = 134
NR_personality = 135
NR_ustat = 136
NR_statfs = 137
NR_fstatfs = 138
NR_sysfs = 139
NR_getpriority = 140
NR_setpriority = 141
NR_sched_setparam = 142
NR_sched_getparam = 143
NR_sched_setscheduler = 144
NR_sched_getscheduler = 145
NR_sched_get_priority_max = 146
NR_sched_get_priority_min = 147
NR_sched_rr_get_interval = 148
NR_mlock = 149
NR_munlock = 150
NR_mlockall = 151
NR_munlockall = 152
NR_vhangup = 153
NR_modify_ldt = 154
NR_pivot_root = 155
NR__sysctl = 156
NR_prctl = 157
NR_arch_prctl = 158
NR_adjtimex = 159
NR_setrlimit = 160
NR_chroot = 161
NR_sync = 162
NR_acct = 163
NR_settimeofday = 164
NR_mount = 165
NR_umount2 = 166
NR_swapon = 167
NR_swapoff = 168
NR_reboot = 169
NR_sethostname = 170
NR_setdomainname = 171
NR_iopl = 172
NR_ioperm = 173
NR_create_module = 174
NR_init_module = 175
NR_delete_module = 176
NR_get_kernel_syms = 177
NR_query_module = 178
NR_quotactl = 179
NR_nfsservctl = 180
NR_getpmsg = 181
NR_putpmsg = 182
NR_afs_syscall = 183
NR_tuxcall = 184
NR_security = 185
NR_gettid = 186
NR_readahead = 187
NR_setxattr = 188
NR_lsetxattr = 189
NR_fsetxattr = 190
NR_getxattr = 191
NR_lgetxattr = 192
NR_fgetxattr = 193
NR_listxattr = 194
NR_llistxattr = 195
NR_flistxattr = 196
NR_removexattr = 197
NR_lremovexattr = 198
NR_fremovexattr = 199
NR_tkill = 200
NR_time = 201
NR_futex = 202
NR_sched_setaffinity = 203
NR_sched_getaffinity = 204
NR_set_thread_area = 205
NR_io_setup = 206
NR_io_destroy = 207
NR_io_getevents = 208
NR_io_submit = 209
NR_io_cancel = 210
NR_get_thread_area = 211
NR_lookup_dcookie = 212
NR_epoll_create = 213
NR_epoll_ctl_old = 214
NR_epoll_wait_old = 215
NR_remap_file_pages = 216
NR_getdents64 = 217
NR_set_tid_address = 218
NR_restart_syscall = 219
NR_semtimedop = 220
NR_fadvise64 = 221
NR_timer_create = 222
NR_timer_settime = 223
NR_timer_gettime = 224
NR_timer_getoverrun = 225
NR_timer_delete = 226
NR_clock_settime = 227
NR_clock_gettime = 228
NR_clock_getres = 229
NR_clock_nanosleep = 230
NR_exit_group = 231
NR_epoll_wait = 232
NR_epoll_ctl = 233
NR_tgkill = 234
NR_utimes = 235
NR_vserver = 236
NR_mbind = 237
NR_set_mempolicy = 238
NR_get_mempolicy = 239
NR_mq_open = 240
NR_mq_unlink = 241
NR_mq_timedsend = 242
NR_mq_timedreceive = 243
NR_mq_notify = 244
NR_mq_getsetattr = 245
NR_kexec_load = 246
NR_waitid = 247
NR_add_key = 248
NR_request_key = 249
NR_keyctl = 250
NR_ioprio_set = 251
NR_ioprio_get = 252
NR_inotify_init = 253
NR_inotify_add_watch = 254
NR_inotify_rm_watch = 255
NR_migrate_pages = 256
NR_openat = 257
NR_mkdirat = 258
NR_mknodat = 259
NR_fchownat = 260
NR_futimesat = 261
NR_newfstatat = 262
NR_unlinkat = 263
NR_renameat = 264
NR_linkat = 265
NR_symlinkat = 266
NR_readlinkat = 267
NR_fchmodat = 268
NR_faccessat = 269
NR_pselect6 = 270
NR_ppoll = 271
NR_unshare = 272
NR_set_robust_list = 273
NR_get_robust_list = 274
NR_splice = 275
NR_tee = 276
NR_sync_file_range = 277
NR_vmsplice = 278
NR_move_pages = 279
NR_utimensat = 280
NR_epoll_pwait = 281
NR_signalfd = 282
NR_timerfd_create = 283
NR_eventfd = 284
NR_fallocate = 285
NR_timerfd_settime = 286
NR_timerfd_gettime = 287
NR_accept4 = 288
NR_signalfd4 = 289
NR_eventfd2 = 290
NR_epoll_create1 = 291
NR_dup3 = 292
NR_pipe2 = 293
NR_inotify_init1 = 294
NR_preadv = 295
NR_pwritev = 296
NR_rt_tgsigqueueinfo = 297
NR_perf_event_open = 298
NR_recvmmsg = 299
NR_fanotify_init = 300
NR_fanotify_mark = 301
NR_prlimit64 = 302
NR_name_to_handle_at = 303
NR_open_by_handle_at = 304
NR_clock_adjtime = 305
NR_syncfs = 306
NR_sendmmsg = 307
NR_setns = 308
NR_getcpu = 309
NR_process_vm_readv = 310
NR_process_vm_writev = 311
NR_kcmp = 312
NR_finit_module = 313
NR_sched_setattr = 314
NR_sched_getattr = 315
NR_renameat2 = 316
NR_seccomp = 317
NR_getrandom = 318
NR_memfd_create = 319
NR_kexec_file_load = 320
NR_bpf = 321
NR_execveat = 322
NR_userfaultfd = 323
NR_membarrier = 324
NR_mlock2 = 325
NR_copy_file_range = 326
NR_preadv2 = 327
NR_pwritev2 = 328
NR_pkey_mprotect = 329
NR_pkey_alloc = 330
NR_pkey_free = 331
NR_statx = 332



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from qiling.os.uefi.ProcessorBind import *

class EFI_GUID(STRUCT):
	_fields_ = [
		('Data1', UINT32),
		('Data2', UINT16),
		('Data3', UINT16),
		('Data4', UINT8 * 8)
	]

EFI_STATUS = UINTN
EFI_HANDLE = PTR(VOID)
EFI_EVENT = PTR(VOID)
EFI_TPL = UINTN
EFI_LBA = UINT64
EFI_PHYSICAL_ADDRESS = UINT64
EFI_VIRTUAL_ADDRESS = UINT64

class EFI_TIME(STRUCT):
	_fields_ = [
		('Year',		UINT16),
		('Month',		UINT8),
		('Day',			UINT8),
		('Hour',		UINT8),
		('Minute',		UINT8),
		('Second',		UINT8),
		('Pad1',		UINT8),
		('Nanosecond',	UINT32),
		('TimeZone',	UINT16),
		('Daylight',	UINT8),
		('Pad2',		UINT8)
	]

__all__ = [
	'EFI_GUID',
	'EFI_STATUS',
	'EFI_HANDLE',
	'EFI_EVENT',
	'EFI_TPL',
	'EFI_LBA',
	'EFI_PHYSICAL_ADDRESS',
	'EFI_VIRTUAL_ADDRESS',
	'EFI_TIME'
]



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import cmd
from functools import partial

from qiling import *
from qiling.const import *
from qiling.debugger import QlDebugger

from .frontend import context_printer, context_reg, context_asm, examine_mem
from .utils import parse_int, handle_bnj, is_thumb, diff_snapshot_save, diff_snapshot_restore, CODE_END
from .const import *


class QlQdb(cmd.Cmd, QlDebugger):
    def __init__(self, ql, init_hook=None, rr=False):

        self._ql = ql
        self.prompt = "(Qdb) "
        self.breakpoints = {}
        self._saved_states = None
        if rr:
            self._states_list = [None]

        super().__init__()

        # setup a breakpoint at entry point or user specified address
        init_hook = self._ql.loader.entry_point if not init_hook else parse_int(init_hook)
        self.set_breakpoint(init_hook, _is_temp=True, _inter=True)


    def parseline(self, line):
        """Parse the line into a command name and a string containing
        the arguments.  Returns a tuple containing (command, args, line).
        'command' and 'args' may be None if the line couldn't be parsed.
        """
        line = line.strip()
        if not line:
            return None, None, line
        elif line[0] == '?':
            line = 'help ' + line[1:]
        elif line.startswith('!'):
            if hasattr(self, 'do_shell'):
                line = 'shell ' + line[1:]
            else:
                return None, None, line
        i, n = 0, len(line)
        while i < n and line[i] in self.identchars: i = i+1
        cmd, arg = line[:i], line[i:].strip()
        return cmd, arg, line


    def interactive(self, *args):
        return self.cmdloop()


    def emptyline(self, *args):
        """
        repeat last command
        """
        _lastcmd = getattr(self, "do_" + self.lastcmd, None)
        if _lastcmd:
            return _lastcmd()


    def del_breakpoint(self, address):
        """
        handle internal breakpoint removing operation
        """
        _bp = self.breakpoints.pop(address, None)
        if _bp:
            _bp["hook"].remove()


    def set_breakpoint(self, address, _is_temp=False, _inter=False):
        """
        handle internal breakpoint adding operation
        """
        _bp_func = partial(self._breakpoint_handler, _is_temp=_is_temp, _inter=_inter)

        _hook = self._ql.hook_address(_bp_func, address)
        self.breakpoints.update({address: {"hook": _hook, "hitted": False, "temp": _is_temp}})

        if _is_temp == False:
            print("Breakpoint at 0x%08x" % address)


    def _breakpoint_handler(self, ql, _is_temp, _inter):
        """
        handle all breakpoints
        """
        _cur_addr = ql.reg.arch_pc

        if _is_temp: # remove temporary breakpoint
            self.del_breakpoint(_cur_addr)
        else:
            if self.breakpoints.get(_cur_addr)["hitted"]:
                return

            print("hit breakpoint at 0x%08x" % _cur_addr)
            self.breakpoints.get(_cur_addr)["hitted"] = True

        self.do_context()
        self._ql.emu_stop()

        if _inter:
            self.interactive()


    def do_context(self, *args):
        """
        show context information for current location
        """
        context_reg(self._ql, self._saved_states)
        context_asm(self._ql, self._ql.reg.arch_pc, 4)


    def do_run(self, *args):
        """
        launch qiling instance
        """

        entry = self._ql.loader.entry_point

        self.run(entry)


    def run(self, address=None):
        """
        handle qiling instance launching
        """

        if address is None:
            return

        # for arm thumb mode
        if self._ql.archtype in (QL_ARCH.ARM, QL_ARCH.ARM_THUMB) and is_thumb(self._ql.reg.cpsr):
            address |= 1

        self._ql.emu_start(address, 0)


    def do_backward(self, *args):

        if getattr(self, "_states_list", None) is None or self._states_list[-1] is None:
            print("there is no way back !!!")
        else:
            print("step backward ~")
            current_state_dicts = self._ql.save(cpu_context=True, mem=True, reg=False, fd=False)
            self._ql.restore(diff_snapshot_restore(current_state_dicts, self._states_list.pop()))
            self.do_context()


    def do_step(self, *args):
        """
        execute one instruction at a time
        """

        if self._ql is None:
            print("The program is not being run.")

        else:
            self._saved_states = dict(filter(lambda d: isinstance(d[0], str), self._ql.reg.save().items()))

            if getattr(self, "_states_list", None) is not None:
                current_state_dicts = self._ql.save(cpu_context=True, mem=True, reg=False, fd=False)
                self._states_list.append(diff_snapshot_save(current_state_dicts, self._states_list[-1]))

            _cur_addr = self._ql.reg.arch_pc

            next_stop = handle_bnj(self._ql, _cur_addr)

            if next_stop is CODE_END:
                return True

            # whether bp placed already
            if self.breakpoints.get(next_stop, None):
                self.breakpoints.get(next_stop)["hitted"] = False

            else:
                self.set_breakpoint(next_stop, _is_temp=True)

            self.run(_cur_addr)


    def do_start(self, *args):
        """
        pause at entry point by setting a temporary breakpoint on it
        """
        entry = self._ql.loader.entry_point  # ld.so
        # entry = self._ql.loader.elf_entry # .text of binary

        if self._ql.archtype in (QL_ARCH.ARM, QL_ARCH.ARM_THUMB) and entry & 1:
            entry -= 1

        if entry not in self.breakpoints.keys():
            self.set_breakpoint(entry, _is_temp=True)

        self.do_run()


    def do_breakpoint(self, address):
        """
        set breakpoint on specific address
        """
        if address:
            baddr = parse_int(address)
            self.set_breakpoint(baddr)


    def do_continue(self, *args):
        """
        continue execution till next breakpoint or the end
        """
        if self._ql is not None:
            _cur_addr = self._ql.reg.arch_pc
            print("continued from 0x%08x" % _cur_addr)

            self.run(_cur_addr)


    def do_examine(self, line):
        """
        Examine memory: x/FMT ADDRESS.
        format letter: o(octal), x(hex), d(decimal), u(unsigned decimal), t(binary), f(float), a(address), i(instruction), c(char), s(string) and z(hex, zero padded on the left)
        size letter: b(byte), h(halfword), w(word), g(giant, 8 bytes)
        e.g. x/4wx 0x41414141 , print 4 word size begin from address 0x41414141 in hex
        """

        _args = line.split()
        DEFAULT_FMT = ('x', 4, 1)

        if line.startswith("/"): # followed by format letter and size letter

            def get_fmt(text):
                def extract_count(t):
                    return "".join([s for s in t if s.isdigit()])

                f, s, c = DEFAULT_FMT
                if extract_count(text):
                    c = int(extract_count(text))

                for char in text.strip(str(c)):
                    if char in SIZE_LETTER.keys():
                        s = SIZE_LETTER.get(char)

                    elif char in FORMAT_LETTER:
                        f = char

                return (f, s, c)

            fmt, addr = line.strip("/").split()
            addr = parse_int(addr)
            fmt = get_fmt(fmt)

        elif len(_args) == 1: # only address
            addr = parse_int(_args[0])
            fmt = DEFAULT_FMT

        else:
            self.do_help("examine")
            return

        try:
            examine_mem(self._ql, addr, fmt)
        except:
            print("something went wrong")


    def do_show(self, *args):
        """
        show some runtime informations
        """
        self._ql.mem.show_mapinfo()
        print("Qdb:", [(hex(idx), val) for idx, val in self.breakpoints.items()])
        print("internal:", [(hex(idx), val) for idx, val in self._ql._addr_hook.items()])


    def do_disassemble(self, address):
        """
        disassemble instructions from address specified
        """
        context_asm(self._ql, parse_int(address), 4)


    def do_shell(self, *command):
        """
        run python code,also a space between exclamation mark and command was necessary
        """
        try:
            print(eval(*command))
        except:
            print("something went wrong")


    def do_quit(self, *args):
        """
        exit Qdb
        """
        return True

    do_r = do_run
    do_s = do_step
    do_q = do_quit
    do_x = do_examine
    do_c = do_continue
    do_b = do_breakpoint
    do_p = do_backward
    do_dis = do_disassemble
    


if __name__ == "__main__":
    pass




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

# https://opensource.apple.com/source/cctools/cctools-795/include/mach-o/loader.h

# magic
# MAGIC_32                = 0xFEEDFACE
# MAGIC_64                = 0xFEEDFACF
# MAGIC_FAT               = 0xBEBAFECA
MAGIC_32                = [0xFEEDFACE, 0xCEFAEDFE]
MAGIC_64                = [0xFEEDFACF, 0xCFFAEDFE]
MAGIC_FAT               = [0xBEBAFECA, 0xCAFEBABE]

# cpu type
CPU_TYPE_X8664          = 0x01000007
CPU_TYPE_ARM64          = 0x0100000C
CPU_TYPE_X86            = 0x00000007
CPU_SUBTYPE_ARM64_ALL   = 0x00000006
CPU_SUBTYPE_X8664_ALL   = 0x00000030
CPU_SUBTYPE_i386_ALL    = 0x00000030

# file type
MH_DYLINKER             = 0x00000007
MH_EXECUTE              = 0x00000002

# load command 
LC_SEGMENT_64           = 0x00000019
LC_SEGMENT              = 0x00000001
LC_SYMTAB               = 0x00000002
LC_DYSYMTAB             = 0x0000000B
LC_ID_DYLINKER          = 0x0000000F
LC_UUID                 = 0x0000001B
LC_VERSION_MIN_MACOSX   = 0x00000024
LC_VERSION_MIN_IPHONEOS = 0x00000025
LC_SOURCE_VERSION       = 0x0000002A
LC_UNIXTHREAD           = 0x00000005
LC_SEGMENT_SPLIT_INFO   = 0x0000001E
LC_FUNCTION_STARTS      = 0x00000026
LC_DATA_IN_CODE         = 0x00000029
LC_CODE_SIGNATURE       = 0x0000001D
LC_DYLD_INFO_ONLY       = 0x80000022
LC_LOAD_DYLINKER        = 0x0000000E
LC_MAIN                 = 0x80000028
LC_LOAD_DYLIB           = 0x0000000C
LC_ENCRYPTION_INFO_64   = 0x0000002C
LC_BUILD_VERSION        = 0x00000032
LC_DYLD_EXPORTS_TRIE    = 0x80000033
LC_DYLD_CHAINED_FIXUPS  = 0x80000034

# UNIXTHREAD 
X86_THREAD_STATE32      = 0x00000001
X86_THREAD_STATE64      = 0x00000004
ARM_THREAD_STATE64      = 0x00000006


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from unicorn import *
from unicorn.x86_const import *

from struct import pack

from .arch import QlArch
from .x86_const import *
from qiling.const import *
from qiling.exception import *


class QlArchX86(QlArch):
    def __init__(self, ql):
        super(QlArchX86, self).__init__(ql)

        x86_register_mappings = [
            reg_map_8, reg_map_16, reg_map_32,
            reg_map_cr, reg_map_st, reg_map_misc
        ]

        for reg_maper in x86_register_mappings:
            self.ql.reg.expand_mapping(reg_maper)

        self.ql.reg.create_reverse_mapping()

        self.ql.reg.register_sp(reg_map_32["esp"])
        self.ql.reg.register_pc(reg_map_32["eip"])


    def stack_push(self, value):
        self.ql.reg.esp -= 4
        self.ql.mem.write(self.ql.reg.esp , self.ql.pack32(value))
        return self.ql.reg.esp


    def stack_pop(self):
        data = self.ql.unpack32(self.ql.mem.read(self.ql.reg.esp, 4))
        self.ql.reg.esp += 4
        return data


    def stack_read(self, offset):
        return self.ql.unpack32(self.ql.mem.read(self.ql.reg.esp+offset, 4))


    def stack_write(self, offset, data):
        return self.ql.mem.write(self.ql.reg.esp + offset, self.ql.pack32(data))


    # get register big, mostly use for x86    
    def get_reg_bit(self, register):
        if type(register) == str:
            register = self.ql.reg.get_uc_reg(register)
        if register in ({v for k, v in reg_map_32.items()}):
            return 32 


    # get initialized unicorn engine
    def get_init_uc(self):
        uc = Uc(UC_ARCH_X86, UC_MODE_32)  
        return uc


class QlArchX8664(QlArch):
    def __init__(self, ql):
        super(QlArchX8664, self).__init__(ql)

        x64_register_mappings = [
            reg_map_8, reg_map_16, reg_map_32, reg_map_64,
            reg_map_cr, reg_map_st, reg_map_misc, reg_map_r, reg_map_seg_base
        ]

        for reg_maper in x64_register_mappings:
            self.ql.reg.expand_mapping(reg_maper)

        self.ql.reg.create_reverse_mapping()

        self.ql.reg.register_sp(reg_map_64["rsp"])
        self.ql.reg.register_pc(reg_map_64["rip"])


    def stack_push(self, value):
        self.ql.reg.rsp -= 8
        self.ql.mem.write(self.ql.reg.rsp, self.ql.pack64(value))
        return self.ql.reg.rsp


    def stack_pop(self):
        data = self.ql.unpack64(self.ql.mem.read(self.ql.reg.rsp, 8))
        self.ql.reg.rsp += 8
        return data


    def stack_read(self, offset):
        return self.ql.unpack64(self.ql.mem.read(self.ql.reg.rsp + offset, 8))


    def stack_write(self, offset, data):
        return self.ql.mem.write(self.ql.reg.rsp + offset, self.ql.pack64(data))


    # get initialized unicorn engine
    def get_init_uc(self):
        uc = Uc(UC_ARCH_X86, UC_MODE_64)  
        return uc


    # get register big, mostly use for x86  
    def get_reg_bit(self, register):
        if type(register) == str:
            register = self.ql.reg.get_uc_reg(register)

        if register in ({v for k, v in reg_map_64.items()}):
            return 64
        else:
            return 32


class GDTManager:
    # Added GDT management module.
    def __init__(self, ql, GDT_ADDR = QL_X86_GDT_ADDR, GDT_LIMIT =  QL_X86_GDT_LIMIT, GDT_ENTRY_ENTRIES = 16):
        ql.log.debug(f"Map GDT at {hex(GDT_ADDR)} with GDT_LIMIT={GDT_LIMIT}")

        if ql.mem.is_mapped(GDT_ADDR, GDT_LIMIT) == False:
            ql.mem.map(GDT_ADDR, GDT_LIMIT, info="[GDT]")

        # setup GDT by writing to GDTR
        ql.reg.write(UC_X86_REG_GDTR, (0, GDT_ADDR, GDT_LIMIT, 0x0))

        self.ql = ql
        self.gdt_number = GDT_ENTRY_ENTRIES
        # self.gdt_used = [False] * GDT_ENTRY_ENTRIES
        self.gdt_addr = GDT_ADDR
        self.gdt_limit = GDT_LIMIT


    def register_gdt_segment(self, index, SEGMENT_ADDR, SEGMENT_SIZE, SPORT, RPORT):
        # FIXME: Temp fix for FS and GS
        if index in (14, 15):
            if self.ql.mem.is_mapped(SEGMENT_ADDR, SEGMENT_ADDR) == False:
                self.ql.mem.map(SEGMENT_ADDR, SEGMENT_ADDR, info="[FS/GS]")

        if index < 0 or index >= self.gdt_number:
            raise QlGDTError("Ql GDT register index error!")
        # create GDT entry, then write GDT entry into GDT table
        gdt_entry = self._create_gdt_entry(SEGMENT_ADDR, SEGMENT_SIZE, SPORT, QL_X86_F_PROT_32)
        self.ql.mem.write(self.gdt_addr + (index << 3), gdt_entry)
        # self.gdt_used[index] = True
        self.ql.log.debug(f"Write to {hex(self.gdt_addr + (index << 3))} for new entry {gdt_entry}")


    def get_gdt_buf(self, start, end):
        return self.ql.mem.read(self.gdt_addr + (start << 3), (end << 3) - (start << 3))


    def set_gdt_buf(self, start, end, buf):
        return self.ql.mem.write(self.gdt_addr + (start << 3), buf[ : (end << 3) - (start << 3)])


    def get_free_idx(self, start = 0, end = -1):
        # The Linux kernel determines whether the segment is empty by judging whether the content in the current GDT segment is 0.
        if end == -1:
            end = self.gdt_number

        idx = -1
        for i in range(start, end):
            if self.ql.unpack64(self.ql.mem.read(self.gdt_addr + (i << 3), 8)) == 0:
                idx = i
                break

        return idx


    def _create_gdt_entry(self, base, limit, access, flags):
        to_ret = limit & 0xffff
        to_ret |= (base & 0xffffff) << 16
        to_ret |= (access & 0xff) << 40
        to_ret |= ((limit >> 16) & 0xf) << 48
        to_ret |= (flags & 0xff) << 52
        to_ret |= ((base >> 24) & 0xff) << 56
        return pack('<Q', to_ret)


    def _create_selector(self, idx, flags):
        to_ret = flags
        to_ret |= idx << 3
        return to_ret


    def create_selector(self, idx, flags):
        return self._create_selector(idx, flags)


def ql_x86_register_cs(self):
    # While debugging the linux kernel segment, the cs segment was found on the third segment of gdt.
    self.gdtm.register_gdt_segment(3, 0, 0xfffff000, QL_X86_A_PRESENT | QL_X86_A_CODE | QL_X86_A_CODE_READABLE | QL_X86_A_PRIV_3 | QL_X86_A_EXEC | QL_X86_A_DIR_CON_BIT, QL_X86_S_GDT | QL_X86_S_PRIV_3)
    self.ql.reg.cs = self.gdtm.create_selector(3, QL_X86_S_GDT | QL_X86_S_PRIV_3)


def ql_x8664_register_cs(self):
    # While debugging the linux kernel segment, the cs segment was found on the sixth segment of gdt.
    self.gdtm.register_gdt_segment(6, 0, 0xfffffffffffff000, QL_X86_A_PRESENT | QL_X86_A_CODE | QL_X86_A_CODE_READABLE | QL_X86_A_PRIV_3 | QL_X86_A_EXEC | QL_X86_A_DIR_CON_BIT, QL_X86_S_GDT | QL_X86_S_PRIV_3)
    self.ql.reg.cs = self.gdtm.create_selector(6, QL_X86_S_GDT | QL_X86_S_PRIV_3)


def ql_x86_register_ds_ss_es(self):
    # TODO : The section permission here should be QL_X86_A_PRIV_3, but I do n’t know why it can only be set to QL_X86_A_PRIV_0.
    # While debugging the Linux kernel segment, I found that the three segments DS, SS, and ES all point to the same location in the GDT table. 
    # This position is the fifth segment table of GDT.
    self.gdtm.register_gdt_segment(5, 0, 0xfffff000, QL_X86_A_PRESENT | QL_X86_A_DATA | QL_X86_A_DATA_WRITABLE | QL_X86_A_PRIV_0 | QL_X86_A_DIR_CON_BIT, QL_X86_S_GDT | QL_X86_S_PRIV_0)
    self.ql.reg.ds = self.gdtm.create_selector(5, QL_X86_S_GDT | QL_X86_S_PRIV_0)
    self.ql.reg.ss = self.gdtm.create_selector(5, QL_X86_S_GDT | QL_X86_S_PRIV_0)
    self.ql.reg.es = self.gdtm.create_selector(5, QL_X86_S_GDT | QL_X86_S_PRIV_0)


def ql_x8664_register_ds_ss_es(self):
    # TODO : The section permission here should be QL_X86_A_PRIV_3, but I do n’t know why it can only be set to QL_X86_A_PRIV_0.
    # When I debug the Linux kernel, I find that only the SS is set to the fifth segment table, and the rest are not set.
    self.gdtm.register_gdt_segment(5, 0, 0xfffff000, QL_X86_A_PRESENT | QL_X86_A_DATA | QL_X86_A_DATA_WRITABLE | QL_X86_A_PRIV_0 | QL_X86_A_DIR_CON_BIT, QL_X86_S_GDT | QL_X86_S_PRIV_0)
    # ql.reg.write(UC_X86_REG_DS, ql.os.gdtm.create_selector(5, QL_X86_S_GDT | QL_X86_S_PRIV_0))
    self.ql.reg.ss = self.gdtm.create_selector(5, QL_X86_S_GDT | QL_X86_S_PRIV_0)
    # ql.reg.write(UC_X86_REG_ES, ql.os.gdtm.create_selector(5, QL_X86_S_GDT | QL_X86_S_PRIV_0))


def ql_x86_register_gs(self):
    self.gdtm.register_gdt_segment(15, GS_SEGMENT_ADDR, GS_SEGMENT_SIZE, QL_X86_A_PRESENT | QL_X86_A_DATA | QL_X86_A_DATA_WRITABLE | QL_X86_A_PRIV_3 | QL_X86_A_DIR_CON_BIT, QL_X86_S_GDT |  QL_X86_S_PRIV_3)
    self.ql.reg.gs = self.gdtm.create_selector(15, QL_X86_S_GDT | QL_X86_S_PRIV_0)


def ql_x86_register_fs(self):
    self.gdtm.register_gdt_segment(14, FS_SEGMENT_ADDR, FS_SEGMENT_SIZE, QL_X86_A_PRESENT | QL_X86_A_DATA | QL_X86_A_DATA_WRITABLE | QL_X86_A_PRIV_3 | QL_X86_A_DIR_CON_BIT, QL_X86_S_GDT |  QL_X86_S_PRIV_3)
    self.ql.reg.fs = self.gdtm.create_selector(14,  QL_X86_S_GDT |  QL_X86_S_PRIV_3)


def ql_x8664_set_gs(ql):
    if ql.mem.is_mapped(GS_SEGMENT_ADDR, GS_SEGMENT_SIZE) == False:
        ql.mem.map(GS_SEGMENT_ADDR, GS_SEGMENT_SIZE, info="[GS]")
    ql.reg.msr(GSMSR, GS_SEGMENT_ADDR)


def ql_x8664_get_gs(ql):
    return ql.reg.msr(GSMSR)


def ql_x8664_set_fs(ql, addr):
    ql.reg.msr(FSMSR, addr)


def ql_x8664_get_fs(ql):
    return ql.reg.msr(FSMSR)



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct
from qiling.os.windows.fncc import *
from qiling.os.const import *
from qiling.os.windows.utils import *


dllname = 'wininet_dll'

# void InternetOpenA(
#   LPCSTR lpszAgent,
#   DWORD  dwAccessType,
#   LPCSTR lpszProxy,
#   LPCSTR lpszProxyBypass,
#   DWORD  dwFlags
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetOpenA(ql, address, params):
    pass


# void InternetOpenW(
#   LPCWSTR lpszAgent,
#   DWORD   dwAccessType,
#   LPCWSTR lpszProxy,
#   LPCWSTR lpszProxyBypass,
#   DWORD   dwFlags
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetOpenW(ql, address, params):
    pass


# void InternetOpenUrlA(
#   HINTERNET hInternet,
#   LPCSTR    lpszUrl,
#   LPCSTR    lpszHeaders,
#   DWORD     dwHeadersLength,
#   DWORD     dwFlags,
#   DWORD_PTR dwContext
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetOpenUrlA(ql, address, params):
    pass


# void InternetOpenUrlW(
#   HINTERNET hInternet,
#   LPCWSTR   lpszUrl,
#   LPCWSTR   lpszHeaders,
#   DWORD     dwHeadersLength,
#   DWORD     dwFlags,
#   DWORD_PTR dwContext
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetOpenUrlW(ql, address, params):
    pass


# BOOLAPI InternetCloseHandle(
#   HINTERNET hInternet
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetCloseHandle(ql, address, params):
    ret = 1
    return ret


# void InternetConnectA(
#   HINTERNET     hInternet,
#   LPCSTR        lpszServerName,
#   INTERNET_PORT nServerPort,
#   LPCSTR        lpszUserName,
#   LPCSTR        lpszPassword,
#   DWORD         dwService,
#   DWORD         dwFlags,
#   DWORD_PTR     dwContext
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetConnectA(ql, address, params):
    pass


# void InternetConnectW(
#   HINTERNET     hInternet,
#   LPCWSTR       lpszServerName,
#   INTERNET_PORT nServerPort,
#   LPCWSTR       lpszUserName,
#   LPCWSTR       lpszPassword,
#   DWORD         dwService,
#   DWORD         dwFlags,
#   DWORD_PTR     dwContext
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetConnectW(ql, address, params):
    pass


# void InternetCheckConnectionA(
#    LPCSTR lpszUrl,
#    DWORD  dwFlags,
#    DWORD  dwReserved
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetCheckConnectionA(ql, address, params):
    pass


# void HttpOpenRequestA(
#   HINTERNET hConnect,
#   LPCSTR    lpszVerb,
#   LPCSTR    lpszObjectName,
#   LPCSTR    lpszVersion,
#   LPCSTR    lpszReferrer,
#   LPCSTR    *lplpszAcceptTypes,
#   DWORD     dwFlags,
#   DWORD_PTR dwContext
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_HttpOpenRequestA(ql, address, params):
    pass


# vBOOLAPI HttpSendRequestExA(
#   HINTERNET           hRequest,
#   LPINTERNET_BUFFERSA lpBuffersIn,
#   LPINTERNET_BUFFERSA lpBuffersOut,
#   DWORD               dwFlags,
#   DWORD_PTR           dwContext
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={'lpBuffersIn': 'POINTER', 'lpBuffersOut': 'STRING'})
def hook_HttpSendRequestExA(ql, address, params):
    pass


# void HttpOpenRequestW(
#   HINTERNET hConnect,
#   LPCWSTR   lpszVerb,
#   LPCWSTR   lpszObjectName,
#   LPCWSTR   lpszVersion,
#   LPCWSTR   lpszReferrer,
#   LPCWSTR   *lplpszAcceptTypes,
#   DWORD     dwFlags,
#   DWORD_PTR dwContext
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_HttpOpenRequestW(ql, address, params):
    pass


# BOOLAPI InternetSetOptionA(
#   HINTERNET hInternet,
#   DWORD     dwOption,
#   LPVOID    lpBuffer,
#   DWORD     dwBufferLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetSetOptionA(ql, address, params):
    return 1


# BOOLAPI InternetSetOptionW(
#   HINTERNET hInternet,
#   DWORD     dwOption,
#   LPVOID    lpBuffer,
#   DWORD     dwBufferLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetSetOptionW(ql, address, params):
    return 1


# BOOLAPI HttpSendRequestA(
#  HINTERNET hRequest,
#  LPCSTR    lpszHeaders,
#  DWORD     dwHeadersLength,
#  LPVOID    lpOptional,
#  DWORD     dwOptionalLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_HttpSendRequestA(ql, address, params):
    return 1


# BOOLAPI HttpSendRequestW(
#  HINTERNET hRequest,
#  LPCWSTR   lpszHeaders,
#  DWORD     dwHeadersLength,
#  LPVOID    lpOptional,
#  DWORD     dwOptionalLength
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_HttpSendRequestW(ql, address, params):
    return 1


# void InternetErrorDlg(
#   HWND      hWnd,
#   HINTERNET hRequest,
#   DWORD     dwError,
#   DWORD     dwFlags,
#   LPVOID    *lppvData
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'HWND': 'POINTER'})
def hook_InternetErrorDlg(ql, address, params):
    pass


# BOOLAPI InternetReadFile(
#   HINTERNET hFile,
#   LPVOID    lpBuffer,
#   DWORD     dwNumberOfBytesToRead,
#   LPDWORD   lpdwNumberOfBytesRead
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetReadFile(ql, address, params):
    return 1


# BOOLAPI InternetWriteFile(
#   HINTERNET hFile,
#   LPCVOID   lpBuffer,
#   DWORD     dwNumberOfBytesToWrite,
#   LPDWORD   lpdwNumberOfBytesWritten
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_InternetWriteFile(ql, address, params):
    return 1


# BOOLAPI HttpEndRequestA(
#   HINTERNET           hRequest,
#   LPINTERNET_BUFFERSA lpBuffersOut,
#   DWORD               dwFlags,
#   DWORD_PTR           dwContext
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPINTERNET_BUFFERSA': 'POINTER'})
def hook_HttpEndRequestA(ql, address, params):
    return 1



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from unicorn import *
from qiling.const import *


class QlGdbUtils(object):
    def __init__(self):
        self.current_address = 0x0
        self.current_address_size = 0x0
        self.last_bp = 0x0
        self.ql = None
        self.entry_point = None
        self.exit_point = None
        self.soft_bp = False
        self.has_soft_bp = False
        self.bp_list = []
        self.mapping = []
        self.breakpoint_count = 0x0
        self.skip_bp_count = 0x0


    def initialize(self, ql, exit_point=None, mappings=None):
        self.ql = ql
        self.current_address = self.entry_point = self.ql.os.entry_point
        self.exit_point = exit_point
        self.mapping = mappings
        self.ql.hook_code(self.dbg_hook)


    def dbg_hook(self, ql, address, size):
        """
        Modified this function for qiling.gdbserver by kabeor from https://github.com/iGio90/uDdbg
        """
        try:
            if self.ql.archtype == QL_ARCH.ARM:
                mode = self.ql.arch.check_thumb()
                if mode == UC_MODE_THUMB:
                    address = address + 1

            self.mapping.append([(hex(address))])
            self.current_address = address
            hit_soft_bp = False

            if self.soft_bp == True:
                self.soft_bp = False
                hit_soft_bp = True

            if (address in self.bp_list and address != self.last_bp) or self.has_soft_bp == True:
                if self.skip_bp_count > 0:
                    self.skip_bp_count -= 1
                else:
                    self.breakpoint_count += 1
                    self.ql.os.stop()
                    self.last_bp = address
                    ql.log.info("gdb> Breakpoint found, stop at address: 0x%x" % address)
                          
            elif address == self.last_bp:
                self.last_bp = 0x0

            self.has_soft_bp = hit_soft_bp
            
            if self.current_address + size == self.exit_point:
                ql.log.debug("gdb> emulation entrypoint at 0x%x" % (self.entry_point))
                ql.log.debug("gdb> emulation exitpoint at 0x%x" % (self.exit_point))
        
        except KeyboardInterrupt as ex:
            ql.log.info("gdb> Paused at 0x%x, instruction size = %u" % (address, size))
            self.ql.os.stop()
        except:
            raise    


    def bp_insert(self, addr):
        if addr not in self.bp_list:
            self.bp_list.append(addr)
            self.ql.log.info('gdb> Breakpoint added at: 0x%x' % addr)


    def bp_remove(self, addr, type = None, len = None):
        self.bp_list.remove(addr)
        self.ql.log.info('gdb> Breakpoint removed at: 0x%x' % addr)


    def resume_emu(self, address=None, skip_bp=0):
        """
        Modified this function for qiling.gdbserver by kabeor from https://github.com/iGio90/uDdbg
        """

        if address is not None:
            if self.ql.archtype == QL_ARCH.ARM:
                mode = self.ql.arch.check_thumb()
                if mode == UC_MODE_THUMB:
                    address += 1
            self.current_address = address

        self.skip_bp_count = skip_bp
        if self.exit_point is not None:
            self.ql.log.info('gdb> Resume at: 0x%x' % self.current_address)
            self.ql.emu_start(self.current_address, self.exit_point)
            


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os
import time

from qiling.os.windows.fncc import *
from qiling.os.const import *
from qiling.os.windows.const import *

dllname = 'msvcrt_dll'

# void __set_app_type (
#    int at
# )
@winsdkapi(cc=CDECL)
def hook___set_app_type(ql, address, params):
    pass


# int __getmainargs(
#     int * _Argc,
#    char *** _Argv,
#    char *** _Env,
#    int _DoWildCard,
# _startupinfo * _StartInfo);
@winsdkapi(cc=CDECL,
    replace_params={"_Argc": POINTER, "_Argv": POINTER, "_Env": POINTER, "_DoWildCard": INT, "_StartInfo": POINTER})
def hook___getmainargs(ql, address, params):
    ret = 0
    return ret


# int* __p__fmode(
# );
@winsdkapi(cc=CDECL)
def hook___p__fmode(ql, address, params):
    addr = ql.os.heap.alloc(ql.pointersize)
    return addr


# int * __p__commode(
#    );
@winsdkapi(cc=CDECL)
def hook___p__commode(ql, address, params):
    addr = ql.os.heap.alloc(ql.pointersize)
    return addr


# char ** __p__acmdln(
#    );
@winsdkapi(cc=CDECL)
def hook___p__acmdln(ql, address, params):
    addr = ql.loader.import_address_table['msvcrt.dll'][b'_acmdln']
    return addr

# wchar_t ** __p__wcmdln(
#    );
@winsdkapi(cc=CDECL)
def hook___p__wcmdln(ql, address, params):
    addr = ql.loader.import_address_table['msvcrt.dll'][b'_wcmdln']
    return addr


# unsigned int _controlfp(
#    unsigned int new,
#    unsigned int mask
# );
@winsdkapi(cc=CDECL, replace_params={"new": UINT, "mask": UINT})
def hook__controlfp(ql, address, params):
    ret = 0x8001f
    return ret


# int atexit(
#    void (__cdecl *func)(void)
# );
@winsdkapi(cc=CDECL, replace_params={"func": POINTER})
def hook_atexit(ql, address, params):
    ret = 0
    return ret


# char*** __p__environ(void)
@winsdkapi(cc=CDECL)
def hook___p__environ(ql, address, params):
    ret = ql.os.heap.alloc(ql.pointersize * len(ql.os.env))
    count = 0
    for key in ql.os.env:
        pointer = ql.os.heap.alloc(ql.pointersize)
        env = key + "=" + ql.os.env[key]
        env_addr = ql.os.heap.alloc(len(env) + 1)
        ql.mem.write(env_addr, bytes(env, 'ascii') + b'\x00')
        ql.mem.write(pointer, ql.pack(env_addr))
        ql.mem.write(ret + count * ql.pointersize, ql.pack(pointer))
        count += 1
    return ret


# int puts(
#    const char *str
# );
@winsdkapi(cc=CDECL, replace_params={"str": STRING})
def hook_puts(ql, address, params):
    ret = 0
    string = params["str"]
    ql.os.stdout.write(bytes(string + "\n", "utf-8"))
    ret = len(string) + 1
    return ret


# void _cexit( void );
@winsdkapi(cc=CDECL)
def hook__cexit(ql, address, params):
    pass


# void __cdecl _initterm(
#    PVFV *,
#    PVFV *
# );
@winsdkapi(cc=CDECL, replace_params={"pfbegin": POINTER, "pfend": POINTER})
def hook__initterm(ql, address, params):
    pass


# void exit(
#    int const status
# );
@winsdkapi(cc=CDECL, replace_params={"status": INT})
def hook_exit(ql, address, params):
    ql.emu_stop()
    ql.os.PE_RUN = False


# int __cdecl _initterm_e(
#    PVFV *,
#    PVFV *
# );
@winsdkapi(cc=CDECL, replace_params={"pfbegin": POINTER, "pfend": POINTER})
def hook__initterm_e(ql, address, params):
    return 0


# char***    __cdecl __p___argv (void);
@winsdkapi(cc=CDECL)
def hook___p___argv(ql, address, params):
    ret = ql.os.heap.alloc(ql.pointersize)
    argv_addr = ql.os.heap.alloc(ql.pointersize * len(ql.os.argv))
    count = 0
    for each in ql.os.argv:
        argv = ql.os.heap.alloc(len(each) + 1)
        ql.mem.write(argv, bytes(each, 'ascii') + b'\x00')
        ql.mem.write(argv_addr + count * ql.pointersize, ql.pack(argv))
        count += 1
    ql.mem.write(ret, ql.pack(argv_addr))
    return ret


# int* __p___argc(void)
@winsdkapi(cc=CDECL)
def hook___p___argc(ql, address, params):
    ql.log.debug("_p___argc")
    ret = ql.os.heap.alloc(ql.pointersize)
    ql.mem.write(ret, ql.pack(len(ql.argv)))
    return ret


@winsdkapi(cc=CDECL)
def hook__get_initial_narrow_environment(ql, address, params):
    ret = 0
    count = 0
    for key in ql.env:
        value = key + "=" + ql.env[key]
        env = ql.os.heap.alloc(len(value) + 1)
        if count == 0:
            ret = env
        ql.mem.write(env, bytes(value, 'ascii') + b'\x00')
        count += 1
    return ret

# int sprintf ( char * str, const char * format, ... );
@winsdkapi(cc=CDECL, dllname=dllname, param_num=3)
def hook_sprintf(ql, address, _):
    ret = 0
    str_ptr, format_ptr = ql.os.get_function_param(2)

    if not format_ptr:
        ql.log.info('printf(format = 0x0) = 0x%x' % ret)
        return ret

    sp = ql.reg.esp if ql.archtype == QL_ARCH.X86 else ql.reg.rsp
    p_args = sp + ql.pointersize * 3

    format_string = ql.os.read_cstring(format_ptr)
    str_size, str_data = ql.os.vprintf(address, format_string, p_args, "sprintf")
    ql.log.info()

    count = format_string.count('%')
    if ql.archtype == QL_ARCH.X8664:
        if count + 1 > 4:
            ql.reg.rsp = ql.reg.rsp + ((count - 4 + 1) * 8)

    ql.mem.write(str_ptr, str_data.encode('utf-8') + b'\x00')
    ret = str_size
    
    return ret


# int printf(const char *format, ...)
@winsdkapi(cc=CDECL, param_num=1)
def hook_printf(ql, address, _):
    ret = 0
    format_string = ql.os.get_function_param(1)

    if format_string == 0:
        ql.log.info('printf(format = 0x0) = 0x%x' % ret)
        return ret

    format_string = ql.os.read_cstring(format_string)

    count = format_string.count("%")
    params = ql.os.get_function_param(count + 1)[1:] if count > 0 else []
    ret, _ = ql.os.printf(address, format_string, params, "printf")

    ql.os.set_return_value(ret)

    # x8664 fastcall donnot known the real number of parameters
    # so you need to manually pop the stack
    if ql.archtype == QL_ARCH.X8664:
        # if number of params > 4
        if count + 1 > 4:
            ql.reg.rsp = ql.reg.rsp + ((count - 4 + 1) * 8)

    return None

# int wprintf(const wchar_t *format, ...)
@winsdkapi(cc=CDECL, param_num=1)
def hook_wprintf(ql, address, _):
    ret = 0
    format_string = ql.os.get_function_param(1)

    if format_string == 0:
        ql.log.info('wprintf(format = 0x0) = 0x%x' % ret)
        return ret

    format_string = ql.os.read_wstring(format_string)

    count = format_string.count("%")
    params = ql.os.get_function_param(count + 1)[1:] if count > 0 else []
    ret, _ = ql.os.printf(address, format_string, params, "wprintf", wstring=True)

    ql.os.set_return_value(ret)

    # x8664 fastcall donnot known the real number of parameters
    # so you need to manually pop the stack
    if ql.archtype == QL_ARCH.X8664:
        # if number of params > 4
        if count + 1 > 4:
            ql.reg.rsp = ql.reg.rsp + ((count - 4 + 1) * 8)

    return None

# MSVCRT_FILE * CDECL MSVCRT___acrt_iob_func(unsigned idx)
@winsdkapi(cc=CDECL, replace_params={"idx": UINT})
def hook___acrt_iob_func(ql, address, params):
    ret = 0
    return ret


@winsdkapi(cc=CDECL, param_num=2)
def hook___stdio_common_vfprintf(ql, address, _):
    ret = 0
    if ql.pointersize == 8:
        _, _, p_format, _, p_args = ql.os.get_function_param(5)
    else:
        _, _, _, p_format, _, p_args = ql.os.get_function_param(6)
    fmt = ql.os.read_cstring(p_format)
    ql.os.vprintf(address, fmt, p_args, '__stdio_common_vfprintf')
    return ret


@winsdkapi(cc=CDECL, param_num=4)
def hook___stdio_common_vfwprintf(ql, address, _):
    ret = 0
    _, _, _, p_format, _, p_args = ql.os.get_function_param(6)
    fmt = ql.os.read_wstring(p_format)

    ql.os.vprintf(address, fmt, p_args, '__stdio_common_vfwprintf', wstring=True)
    return ret


@winsdkapi(cc=CDECL, param_num=4)
def hook___stdio_common_vswprintf_s(ql, address, _):
    ret = 0
    _, size, p_format, p_args = ql.os.get_function_param(4)

    fmt = ql.os.read_wstring(p_format)
    ql.os.vprintf(address, fmt, p_args, '__stdio_common_vswprintf_s', wstring=True)

    return ret

# int lstrlenA(
#   LPCSTR lpString
# );
@winsdkapi(cc=STDCALL, replace_params={'lpString': POINTER})
def hook_lstrlenA(ql, address, params):
    addr = params["lpString"]

    if addr == 0:
        return 0

    string = b""
    val = ql.mem.read(addr, 1)
    while bytes(val) != b"\x00":
        addr += 1
        string += bytes(val)
        val = ql.mem.read(addr, 1)
    params["lpString"] = bytearray(string)
    return len(string)


# int lstrlenW(
#   LPCWSTR lpString
# );
@winsdkapi(cc=CDECL, replace_params={'lpString': POINTER})
def hook_lstrlenW(ql, address, params):
    addr = params["lpString"]

    if addr == 0:
        return 0

    string = b""
    val = ql.mem.read(addr, 2)
    while bytes(val) != b"\x00\x00":
        addr += 2
        string += bytes(val)
        val = ql.mem.read(addr, 2)
    params["lpString"] = bytearray(string)
    return len(string)


@winsdkapi(cc=CDECL)
def hook___lconv_init(ql, address, params):
    ret = 0
    return ret


# size_t strlen(
#    const char *str
# );
@winsdkapi(cc=CDECL, replace_params={"str": STRING})
def hook_strlen(ql, address, params):
    _str = params["str"]
    strlen = len(_str)
    return strlen


# int strncmp(
#    const char *string1,
#    const char *string2,
#    size_t count
# );
@winsdkapi(cc=CDECL, replace_params={"string1": STRING, "string2": STRING, "count": SIZE_T})
def hook_strncmp(ql, address, params):
    s1 = params["string1"]
    s2 = params["string2"]
    count = params["count"]
    string1 = s1[:count]
    string2 = s2[:count]

    if string1 == string2:
        result = 0
    elif string1 > string2:
        result = 1
    else:
        result = -1
    return result


# void* malloc（unsigned int size)
@winsdkapi(cc=CDECL, replace_params={"size": UINT})
def hook_malloc(ql, address, params):
    size = params['size']
    addr = ql.os.heap.alloc(size)
    return addr


# _onexit_t _onexit(
#    _onexit_t function
# );
@winsdkapi(cc=CDECL, replace_params={"function": POINTER})
def hook__onexit(ql, address, params):
    function = params['function']
    addr = ql.os.heap.alloc(ql.pointersize)
    ql.mem.write(addr, ql.pack(function))
    return addr


# void *memset(
#    void *dest,
#    int c,
#    size_t count
# );
@winsdkapi(cc=CDECL, replace_params={"dest": POINTER, "c": INT, "count": SIZE_T})
def hook_memset(ql, address, params):
    dest = params["dest"]
    c = params["c"]
    count = params["count"]
    ql.mem.write(dest, bytes(c) * count)
    return dest


# void *calloc(
#    size_t num,
#    size_t size
# );
@winsdkapi(cc=CDECL, replace_params={"num": SIZE_T, "size": SIZE_T})
def hook_calloc(ql, address, params):
    num = params['num']
    size = params['size']
    ret = ql.os.heap.alloc(num * size)
    return ret


# void * memmove(
#   void *dest,
#   const void *src,
#   size_t num
# );
@winsdkapi(cc=CDECL, replace_params={"dest": POINTER, "src": POINTER, "num": SIZE_T})
def hook_memmove(ql, address, params):
    data = ql.mem.read(params['src'], params['num'])
    ql.mem.write(params['dest'], bytes(data))
    return params['dest']


# int _ismbblead(
#    unsigned int c
# );
@winsdkapi(cc=CDECL, replace_params={"c": UINT})
def hook__ismbblead(ql, address, params):
    # TODO check if is CDECL or not
    # If locale is utf-8 always return 0
    loc = LOCALE["default"]
    if loc[0x1004] == "utf-8":
        return 0
    else:
        raise QlErrorNotImplemented("API not implemented")


# errno_t _wfopen_s(
#    FILE** pFile,
#    const wchar_t *filename,
#    const wchar_t *mode
# );
@winsdkapi(cc=CDECL, replace_params={"pFile": POINTER, "filename": WSTRING, "mode": WSTRING})
def hook__wfopen_s(ql, address, params):
    dst = params["pFile"]
    filename = params["filename"]
    mode = params["mode"]
    f = ql.os.fs_mapper.open(filename, mode)
    new_handle = Handle(obj=f)
    ql.os.handle_manager.append(new_handle)
    ql.mem.write(dst, ql.pack(new_handle.id))
    return 1


# time_t time( time_t *destTime );
@winsdkapi(cc=CDECL, replace_params={"destTime": POINTER})
def hook__time64(ql, address, params):
    dst = params["destTime"]
    time_wasted = int(time.time())
    if dst != 0:
        ql.mem.write(dst, time_wasted.to_bytes(8, "little"))
    return time_wasted
